# -*- coding: utf-8 -*-
#
#       ramstk.models.stakeholder.table.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Stakeholder Table Model."""

# Standard Library Imports
from typing import Any, Dict, Type

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.analyses import improvementfactor
from ramstk.models import RAMSTKBaseTable, RAMSTKStakeholderRecord


class RAMSTKStakeholderTable(RAMSTKBaseTable):
    """Contain the attributes and methods of the Stakeholder data manager."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _db_id_colname = "fld_stakeholder_id"
    _db_tablename = "ramstk_stakeholder"
    _select_msg = "selected_revision"
    _tag = "stakeholder"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self, **kwargs: Dict[Any, Any]) -> None:
        """Initialize a Stakeholder data manager instance."""
        super().__init__(**kwargs)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_id_columns = [
            "revision_id",
            "stakeholder_id",
            "parent_id",
            "record_id",
        ]

        # Initialize private scalar attributes.
        self._record: Type[RAMSTKStakeholderRecord] = RAMSTKStakeholderRecord

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.pkey = "stakeholder_id"

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_calculate_stakeholder, "request_calculate_stakeholder")

    def do_get_new_record(  # pylint: disable=method-hidden
        self, attributes: Dict[str, Any]
    ) -> object:
        """Gets a new record instance with attributes set.

        :param attributes: the dict of attribute values to assign to the new record.
        :return: None
        :rtype: None
        """
        _new_record = self._record()
        _new_record.revision_id = attributes["revision_id"]
        _new_record.stakeholder_id = self.last_id + 1
        _new_record.description = "New Stakeholder Input"

        return _new_record

    def do_calculate_stakeholder(self, node_id: int) -> None:
        """Calculate improvement factor and weight for currently selected item.

        :param node_id: the ID of the record to calculate.
        :return: None
        :rtype: None
        """
        self._do_calculate_improvement(node_id)

        pub.sendMessage(
            "succeed_calculate_stakeholder",
            tree=self.tree,
        )

    def _do_calculate_improvement(self, node_id: int) -> None:
        """Calculate improvement factor and weight for currently selected item.

        :param node_id: the ID of the record to calculate.
        :return: None
        :rtype: None
        """
        _record = self.tree.get_node(node_id).data[self._tag]
        _attributes = _record.get_attributes()

        (_improvement, _overall_weight,) = improvementfactor.calculate_improvement(
            _attributes["planned_rank"],
            _attributes["customer_rank"],
            _attributes["priority"],
            user_float_1=_attributes["user_float_1"],
            user_float_2=_attributes["user_float_2"],
            user_float_3=_attributes["user_float_3"],
            user_float_4=_attributes["user_float_4"],
            user_float_5=_attributes["user_float_5"],
        )

        self.do_set_attributes(
            node_id=[node_id],
            package={"improvement": _improvement},
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={"overall_weight": _overall_weight},
        )



# -*- coding: utf-8 -*-
#
#       ramstk.models.usage_profile.view.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Usage Profile Package View Model."""

# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from pubsub import pub
from treelib import Tree

# RAMSTK Package Imports
from ramstk.models import RAMSTKBaseView


class RAMSTKUsageProfileView(RAMSTKBaseView):
    """Contain the attributes and methods of the Usage Profile view.

    This class manages the usage profile data from the RAMSTKMissionRecord,
    RAMSTKMissionPhase, and RAMSKTEnvironment table models.
    """

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _root = 0
    _tag = "usage_profile"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self, **kwargs: Dict[Any, Any]) -> None:
        """Initialize a usage profile view model instance."""
        super().__init__(**kwargs)

        # Initialize private dictionary attributes.
        self._dic_load_functions = {
            "mission": self._do_load_missions,
            "mission_phase": self._do_load_mission_phases,
            "environment": self._do_load_environments,
        }
        self._dic_trees = {
            "mission": Tree(),
            "mission_phase": Tree(),
            "environment": Tree(),
        }

        # Initialize private list attributes.
        self._lst_modules = [
            "mission",
            "mission_phase",
            "environment",
        ]

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        # Subscribe to PyPubSub messages.
        pub.subscribe(super().do_set_tree, "succeed_insert_environment")
        pub.subscribe(super().do_set_tree, "succeed_insert_mission")
        pub.subscribe(super().do_set_tree, "succeed_insert_mission_phase")
        pub.subscribe(super().do_set_tree, "succeed_retrieve_environments")
        pub.subscribe(super().do_set_tree, "succeed_retrieve_missions")
        pub.subscribe(super().do_set_tree, "succeed_retrieve_mission_phases")
        pub.subscribe(super().do_set_tree, "succeed_delete_environment")
        pub.subscribe(super().do_set_tree, "succeed_delete_mission")
        pub.subscribe(super().do_set_tree, "succeed_delete_mission_phase")

    def _do_load_environments(self, phase_id: int, parent_id: str) -> None:
        """Load the environments into the tree for the passed phase ID.

        :param phase_id: the mission phase ID to load the environments for.
        :param parent_id: the parent node ID.
        :return: None
        :rtype: None
        """
        for _node in self._dic_trees["environment"].all_nodes()[1:]:
            _environment = _node.data["environment"]
            _node_id = f"{parent_id}.{_environment.environment_id}"

            if _environment.phase_id == phase_id:
                self.tree.create_node(
                    tag="environment",
                    identifier=_node_id,
                    parent=parent_id,
                    data={"usage_profile": _environment},
                )

    def _do_load_missions(self) -> None:
        """Load the missions into the tree for the passed mission ID.

        :return: None
        :rtype: None
        """
        for _node in self._dic_trees["mission"].all_nodes()[1:]:
            _mission = _node.data["mission"]
            _node_id = f"{_mission.mission_id}"

            self.tree.create_node(
                tag="mission",
                identifier=_node_id,
                parent=self._root,
                data={"usage_profile": _mission},
            )

            if self._dic_trees["mission_phase"].depth() > 0:
                self._dic_load_functions["mission_phase"](  # type: ignore
                    _mission.mission_id,
                )

    def _do_load_mission_phases(self, mission_id: int) -> None:
        """Load the mission phases into the tree for the passed mission ID.

        :param mission_id: the mission ID to add the new mission phase.
        :return: None
        :rtype: None
        """
        for _node in self._dic_trees["mission_phase"].all_nodes()[1:]:
            _mission_phase = _node.data["mission_phase"]
            _node_id = f"{mission_id}.{_mission_phase.phase_id}"

            if _mission_phase.mission_id == mission_id:
                self.tree.create_node(
                    tag="mission_phase",
                    identifier=_node_id,
                    parent=f"{mission_id}",
                    data={"usage_profile": _mission_phase},
                )

                if self._dic_trees["environment"].depth() > 0:
                    self._dic_load_functions["environment"](  # type: ignore
                        _mission_phase.phase_id,
                        _node_id,
                    )



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.programdb.RAMSTKProgramImfo.py is part of The RAMSTK
#       Project
#
# All rights reserved.
# Copyright 2007 - 2021 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKProgramInfo Table Module."""

# Standard Library Imports
from datetime import date

# Third Party Imports
from sqlalchemy import Column, Date, Integer, String

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKProgramInfoRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent table ramstk_program_info in RAMSTK Program database.

    This table has a one-to-one relationship with RAMSTKProgramStatus.
    """

    __defaults__ = {
        "function_active": 1,
        "requirement_active": 1,
        "hardware_active": 1,
        "software_active": 0,
        "rcm_active": 0,
        "testing_active": 0,
        "incident_active": 0,
        "survival_active": 0,
        "vandv_active": 1,
        "hazard_active": 1,
        "stakeholder_active": 1,
        "allocation_active": 1,
        "similar_item_active": 1,
        "fmea_active": 1,
        "pof_active": 1,
        "rbd_active": 0,
        "fta_active": 0,
        "created_on": date.today(),
        "created_by": "",
        "last_saved": date.today(),
        "last_saved_by": "",
    }
    __tablename__ = "ramstk_program_info"
    __table_args__ = {"extend_existing": True}

    revision_id = Column(
        "fld_revision_id", Integer, primary_key=True, autoincrement=True, nullable=False
    )
    function_active = Column(
        "fld_function_active", Integer, default=__defaults__["function_active"]
    )
    requirement_active = Column(
        "fld_requirement_active", Integer, default=__defaults__["requirement_active"]
    )
    hardware_active = Column(
        "fld_hardware_active", Integer, default=__defaults__["hardware_active"]
    )
    software_active = Column(
        "fld_software_active", Integer, default=__defaults__["software_active"]
    )
    rcm_active = Column("fld_rcm_active", Integer, default=__defaults__["rcm_active"])
    testing_active = Column(
        "fld_testing_active", Integer, default=__defaults__["testing_active"]
    )
    incident_active = Column(
        "fld_incident_active", Integer, default=__defaults__["incident_active"]
    )
    survival_active = Column(
        "fld_survival_active", Integer, default=__defaults__["survival_active"]
    )
    vandv_active = Column(
        "fld_vandv_active", Integer, default=__defaults__["vandv_active"]
    )
    hazard_active = Column(
        "fld_hazard_active", Integer, default=__defaults__["hazard_active"]
    )
    stakeholder_active = Column(
        "fld_stakeholder_active", Integer, default=__defaults__["stakeholder_active"]
    )
    allocation_active = Column(
        "fld_allocation_active", Integer, default=__defaults__["allocation_active"]
    )
    similar_item_active = Column(
        "fld_similar_item_active", Integer, default=__defaults__["similar_item_active"]
    )
    fmea_active = Column(
        "fld_fmea_active", Integer, default=__defaults__["fmea_active"]
    )
    pof_active = Column("fld_pof_active", Integer, default=__defaults__["pof_active"])
    rbd_active = Column("fld_rbd_active", Integer, default=__defaults__["rbd_active"])
    fta_active = Column("fld_fta_active", Integer, default=__defaults__["fta_active"])
    created_on = Column("fld_created_on", Date, default=__defaults__["created_on"])
    created_by = Column(
        "fld_created_by", String(512), default=__defaults__["created_by"]
    )
    last_saved = Column("fld_last_saved_on", Date, default=__defaults__["last_saved"])
    last_saved_by = Column(
        "fld_last_saved_by", String(512), default=__defaults__["last_saved_by"]
    )

    def get_attributes(self):
        """Retrieve current values of RAMSTKProgramInfo data model attributes.

        :return: {revision_id, function_active, requirement_active,
                  hardware_active, software_active, vandv_active,
                  testing_active, fraca_active, survival_active, rcm_active,
                  rbd_active, fta_active, created_on, created_by, last_saved,
                  last_saved_by, method} pairs.
        :rtype: dict
        """
        _attributes = {
            "revision_id": self.revision_id,
            "function_active": self.function_active,
            "requirement_active": self.requirement_active,
            "hardware_active": self.hardware_active,
            "software_active": self.software_active,
            "rcm_active": self.rcm_active,
            "testing_active": self.testing_active,
            "incident_active": self.incident_active,
            "survival_active": self.survival_active,
            "vandv_active": self.vandv_active,
            "fmea_active": self.fmea_active,
            "pof_active": self.pof_active,
            "hazard_active": self.hazard_active,
            "stakeholder_active": self.stakeholder_active,
            "allocation_active": self.allocation_active,
            "similar_item_active": self.similar_item_active,
            "rbd_active": self.rbd_active,
            "fta_active": self.fta_active,
            "created_on": self.created_on,
            "created_by": self.created_by,
            "last_saved": self.last_saved,
            "last_saved_by": self.last_saved_by,
        }

        return _attributes



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.usage_profile.panel.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Usage Profile Panels."""

# Standard Library Imports
from typing import Any, Dict, List, Tuple

# Third Party Imports
import treelib
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import GdkPixbuf, Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKTreePanel


class UsageProfileTreePanel(RAMSTKTreePanel):
    """Panel to display hierarchical list of usage profiles."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_retrieve_usage_profile"
    _tag = "usage_profile"
    _title = _("Usage Profile")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the usage profile panel."""
        super().__init__()

        # Initialize private dictionary class attributes.
        self.tvwTreeView.dic_row_loader = {
            "mission": self.__do_load_mission,
            "mission_phase": self.__do_load_phase,
            "environment": self.__do_load_environment,
        }
        self._dic_visible_mask: Dict[str, Dict[str, bool]] = {
            "mission": {
                "revision_id": False,
                "mission_id": True,
                "phase_id": False,
                "environment_id": False,
                "name": False,
                "description": True,
                "mission_time": True,
                "units": True,
                "phase_start": False,
                "phase_end": False,
                "minimum": False,
                "maximum": False,
                "mean": False,
                "variance": False,
            },
            "mission_phase": {
                "revision_id": False,
                "mission_id": False,
                "phase_id": True,
                "environment_id": False,
                "name": True,
                "description": True,
                "mission_time": False,
                "units": False,
                "phase_start": True,
                "phase_end": True,
                "minimum": False,
                "maximum": False,
                "mean": False,
                "variance": False,
            },
            "environment": {
                "revision_id": False,
                "mission_id": False,
                "phase_id": False,
                "environment_id": True,
                "name": True,
                "description": False,
                "mission_time": False,
                "units": True,
                "phase_start": False,
                "phase_end": False,
                "minimum": True,
                "maximum": True,
                "mean": True,
                "variance": True,
            },
        }

        # Initialize private list class attributes.

        # Initialize private scalar class attributes.
        self._on_edit_message: str = f"lvw_editing_{self._tag}"

        # Initialize public dictionary class attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "revision_id": [
                0,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Revision ID"),
                "gint",
            ],
            "mission_id": [
                1,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Mission ID"),
                "gint",
            ],
            "phase_id": [
                2,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Phase ID"),
                "gint",
            ],
            "environment_id": [
                3,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Environment ID"),
                "gint",
            ],
            "name": [
                4,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Mission Phase Name"),
                "gchararray",
            ],
            "description": [
                5,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Description"),
                "gchararray",
            ],
            "mission_time": [
                6,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Mission Time"),
                "gfloat",
            ],
            "units": [
                7,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Units"),
                "gchararray",
            ],
            "phase_start": [
                8,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Phase Start"),
                "gfloat",
            ],
            "phase_end": [
                9,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Phase End"),
                "gfloat",
            ],
            "minimum": [
                10,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Minimum"),
                "gfloat",
            ],
            "maximum": [
                11,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Maximum"),
                "gfloat",
            ],
            "mean": [
                12,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Mean"),
                "gfloat",
            ],
            "variance": [
                13,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Variance"),
                "gfloat",
            ],
        }

        self.dic_icons: Dict[str, Any] = {
            "mission": None,
            "mission_phase": None,
            "environment": None,
        }
        self.dic_units: Dict[str, Tuple[str, str, str]] = {}

        # Initialize public list class attributes.

        # Initialize public scalar class attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.tvwTreeView.set_tooltip_text(
            _("Displays the usage profiles for the selected revision.")
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(super()._do_set_attributes, f"lvw_editing_{self._tag}")

    def do_load_comboboxes(self) -> None:
        """Load the Gtk.CellRendererCombo()s.

        :return: None
        :rtype: None
        """
        _model = self.tvwTreeView.get_cell_model(self.tvwTreeView.position["units"])
        for __, _unit in self.dic_units.items():  # pylint: disable=unused-variable
            _model.append([_unit[1]])

    def _on_row_change(self, selection: Gtk.TreeSelection) -> None:
        """Handle row changes for the Usage Profile package List View.

        This method is called whenever a Usage Profile List View
        RAMSTKTreeView() row is activated or changed.

        :param selection: the Usage Profile class Gtk.TreeSelection().
        :return: None
        """
        _attributes = super().on_row_change(selection)
        __, _row = selection.get_selected()  # pylint: disable=unused-variable

        if _row is not None:
            if (
                _attributes["mission_id"]
                * _attributes["phase_id"]
                * _attributes["environment_id"]
            ) > 0:
                self._tag = "environment"
                self._record_id = _attributes["environment_id"]
            elif (_attributes["mission_id"] * _attributes["phase_id"]) > 0:
                self._tag = "mission_phase"
                self._record_id = _attributes["phase_id"]
            else:
                self._tag = "mission"
                self._record_id = _attributes["mission_id"]

            self.tvwTreeView.visible = self._dic_visible_mask[self._tag]
            self.tvwTreeView.do_set_visible_columns()

            pub.sendMessage(f"selected_{self._tag}", attributes=_attributes)

    def __do_load_environment(
        self, node: treelib.Node, row: Gtk.TreeIter
    ) -> Gtk.TreeIter:
        """Load an environmental condition into the RAMSTK TreeView.

        :param node: the treelib Node() with the mode data to load.
        :param row: the parent row of the mode to load into the FMEA form.
        :return: _new_row; the row that was just populated with mode data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        _model = self.tvwTreeView.get_model()

        _pixbuf = GdkPixbuf.Pixbuf()
        _icon = _pixbuf.new_from_file_at_size(self.dic_icons["environment"], 22, 22)

        _attributes = [
            _entity.revision_id,
            _entity.mission_id,
            _entity.phase_id,
            _entity.environment_id,
            _entity.name,
            "",
            0.0,
            _entity.units,
            0.0,
            0.0,
            _entity.minimum,
            _entity.maximum,
            _entity.mean,
            _entity.variance,
            _icon,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                "An error occurred when loading environment {0:s} in the "
                "usage profile.  This might indicate it was missing it's data "
                "package, some of the data in the package was missing, or "
                "some of the data was the wrong type.  Row data was: "
                "{1}"
            ).format(str(node.identifier), _attributes)
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_mission(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a mission into the RAMSTKTreeView().

        :param node: the treelib Node() with the mode data to load.
        :param row: the parent row of the mode to load into the FMEA form.
        :return: _new_row; the row that was just populated with mode data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        # pylint: disable=unused-variable
        [[__, _entity]] = node.data.items()

        _model = self.tvwTreeView.get_model()

        _pixbuf = GdkPixbuf.Pixbuf()
        _icon = _pixbuf.new_from_file_at_size(self.dic_icons["mission"], 22, 22)

        _attributes = [
            _entity.revision_id,
            _entity.mission_id,
            0,
            0,
            "",
            _entity.description,
            _entity.mission_time,
            _entity.time_units,
            0.0,
            _entity.mission_time,
            0.0,
            0.0,
            0.0,
            0.0,
            _icon,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                "An error occurred when loading mission {0:s} in the usage "
                "profile.  This might indicate it was missing it's data "
                "package, some of the data in the package was missing, or "
                "some of the data was the wrong type.  Row data was: "
                "{1}"
            ).format(str(node.identifier), _attributes)
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_phase(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a mission phase into the RAMSTKTreeView().

        :param node: the treelib Node() with the mode data to load.
        :param row: the parent row of the mode to load into the FMEA form.
        :return: _new_row; the row that was just populated with mode data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        _model = self.tvwTreeView.get_model()

        _pixbuf = GdkPixbuf.Pixbuf()
        _icon = _pixbuf.new_from_file_at_size(self.dic_icons["mission_phase"], 22, 22)

        _attributes = [
            _entity.revision_id,
            _entity.mission_id,
            _entity.phase_id,
            0,
            _entity.name,
            _entity.description,
            0.0,
            "",
            _entity.phase_start,
            _entity.phase_end,
            0.0,
            0.0,
            0.0,
            0.0,
            _icon,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                "An error occurred when loading mission phase {0:s} in the "
                "usage profile.  This might indicate it was missing it's data "
                "package, some of the data in the package was missing, or "
                "some of the data was the wrong type.  Row data was: "
                "{1}"
            ).format(str(node.identifier), _attributes)
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.hardware.components.switch.py is part of the RAMSTK
#       Project.
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Switch Results Panel."""

# Standard Library Imports
from typing import Any, Dict, List

# RAMSTK Package Imports
from ramstk.views.gtk3 import _
from ramstk.views.gtk3.milhdbk217f import MilHdbk217FResultPanel
from ramstk.views.gtk3.widgets import RAMSTKEntry


class SwitchMilHdbk217FResultPanel(MilHdbk217FResultPanel):
    """Display Switch assessment results attribute data.

    The Switch assessment result view displays all the assessment results
    for the selected switch.  This includes, currently, results for
    MIL-HDBK-217FN2 parts count and MIL-HDBK-217FN2 part stress methods.  The
    attributes of a switch assessment result view are:

    :ivar txtPiC: displays the contact form and quantity factor for the switch.
    :ivar txtPiCYC: displays the cycling factor for the switch.
    :ivar txtPiL: displays the load stress factor for the switch.
    :ivar txtPiN: displays the number of active contacts factor for the switch.
    :ivar txtPiU: displays the use factor for the breaker.
    """

    # Define private dict class attributes.
    _dic_part_stress: Dict[int, str] = {
        1: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>CYC</sub>\u03C0<sub>L</sub>\u03C0<sub"
        ">C</sub>\u03C0<sub>E</sub></span>",
        2: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>CYC</sub>\u03C0<sub>L</sub>\u03C0<sub"
        ">E</sub></span>",
        3: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>CYC</sub>\u03C0<sub>L</sub>\u03C0<sub"
        ">E</sub></span>",
        4: '<span foreground="blue">\u03BB<sub>p</sub> = (\u03BB<sub>b1</sub> '
        "+ \u03C0<sub>N</sub>\u03BB<sub>b2</sub>)\u03C0<sub>CYC</sub>\u03C0"
        "<sub>L</sub>\u03C0<sub>E</sub></span>",
        5: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>C</sub>\u03C0<sub>U</sub>\u03C0<sub>Q"
        "</sub>\u03C0<sub>E</sub></span> ",
    }

    # Define private list class attributes.

    # Define private scalar class attributes.
    _title: str = _("Switch MIL-HDBK-217F Results")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Switch assessment result view."""
        super().__init__()

        # Initialize widgets.
        self.txtPiC: RAMSTKEntry = RAMSTKEntry()
        self.txtPiCYC: RAMSTKEntry = RAMSTKEntry()
        self.txtPiL: RAMSTKEntry = RAMSTKEntry()
        self.txtPiN: RAMSTKEntry = RAMSTKEntry()
        self.txtPiU: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "hazard_rate_model": [
                13,
                self.lblModel,
                "",
                None,
                "",
                "",
                {
                    "tooltip": _(
                        "The assessment model used to calculate the switch hazard rate."
                    ),
                },
                "",
            ],
            "lambda_b": [
                23,
                self.txtLambdaB,
                "",
                None,
                "",
                0.0,
                {
                    "tooltip": _("The base hazard rate for the switch."),
                },
                "\u03BB<sub>b</sub>:",
            ],
            "pi_q": [
                30,
                self.txtPiQ,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The quality factor for the switch."),
                },
                "\u03C0<sub>Q</sub>:",
            ],
            "pi_e": [
                19,
                self.txtPiE,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The environment factor for the switch."),
                },
                "\u03C0<sub>E</sub>:",
            ],
            "pi_cyc": [
                18,
                self.txtPiCYC,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The cycling factor for the switch."),
                },
                "\u03C0<sub>CYC</sub>:",
            ],
            "pi_l": [
                23,
                self.txtPiL,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The load stress factor for the switch."),
                },
                "\u03C0<sub>L</sub>:",
            ],
            "pi_c": [
                13,
                self.txtPiC,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _(
                        "The number of active contacts factor for the switch."
                    ),
                },
                "\u03C0<sub>C</sub>:",
            ],
            "pi_n": [
                26,
                self.txtPiN,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _(
                        "The contact form and quantity factor for the switch.  This is "
                        "the configuration factor for a circuit breaker."
                    ),
                },
                "\u03C0<sub>N</sub>:",
            ],
            "pi_u": [
                35,
                self.txtPiU,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The use factor for the switch."),
                },
                "\u03C0<sub>U</sub>:",
            ],
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        super().do_set_properties()
        super().do_make_panel()

        # Subscribe to PyPubSub messages.

    def _do_load_entries(self, attributes: Dict[str, Any]) -> None:
        """Load the switch assessment results page.

        :return: None
        :rtype: None
        """
        super().do_load_entries(attributes)

        self.txtPiCYC.set_sensitive(False)
        self.txtPiL.set_sensitive(False)
        self.txtPiC.set_sensitive(False)
        self.txtPiN.set_sensitive(False)
        self.txtPiU.set_sensitive(False)

        if self.category_id == 7 and self._hazard_rate_method_id == 2:
            self.txtPiCYC.do_update(str(self.fmt.format(attributes["piCYC"])))
            self.txtPiL.do_update(str(self.fmt.format(attributes["piL"])))
            self.txtPiC.do_update(str(self.fmt.format(attributes["piC"])))
            self.txtPiN.do_update(str(self.fmt.format(attributes["piN"])))
            self.txtPiU.do_update(str(self.fmt.format(attributes["piU"])))



# type: ignore
# -*- coding: utf-8 -*-
#
#       ramstk.analyses.statistics.lognormal.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Exponential Module."""

# Standard Library Imports
from typing import Tuple

# Third Party Imports
import scipy
from scipy.stats import norm


def get_hazard_rate(location: float, scale: float, time: float) -> float:
    """Calculates the hazard rate given a location and scale parameter.

    This function calculates the rate parameter given the scale, shape, time, and,
    optionally, a location parameter.

        >>> get_hazard_rate(100.0, 10.0, 85.0)
        0.013878975045885079

        >>> get_hazard_rate(0.0, 10.0, 85.0)
        0.8614595320165236

        >>> get_hazard_rate(100.0, 0.0, 85.0)
        0.0

    :param location: the value of the location (mu) parameter.
    :param scale: the value of the scale (sigma) parameter.
    :param time: the time at which to calculate the hazard rate.
    :return: _hazard_rate; the hazard rate.
    :rtype: float
    """
    _hazard_rate = norm.pdf(time, loc=location, scale=scale,) / norm.sf(
        time,
        loc=location,
        scale=scale,
    )

    return _hazard_rate


def get_mtbf(location: float, scale: float) -> float:
    """Calculate the MTBF given a shape (sigma) and scale (mu) parameter.

        >>> get_mtbf(100.0, 10.0)
        100.0

        >>> get_mtbf(0.0, 10.0)
        0.0

        >>> get_mtbf(100.0, 0.0)
        nan

    :param location: the location parameter.
    :param scale: the scale parameter.
    :return: _mtbf; the MTBF.
    :rtype: float
    """
    return norm.mean(
        loc=location,
        scale=scale,
    )


def get_survival(
    location: float,
    scale: float,
    time: float,
) -> float:
    """Calculate value of the survival function at time given scale and location.

        >>> get_survival(100.0, 10.0, 85.0)
        0.9331927987311419

        >>> get_survival(0.0, 10.0, 85.0)
        9.47953482220325e-18

        >>> get_survival(100.0, 0.0, 85.0)
        0.9928813

    :param location: the point estimate of the location parameter.
    :param scale: the point estimate of the scale parameter.
    :param time: the time at which to calculate the survival function.
    :return: _surv; the value of the survival function at time.
    :rtype: float
    """
    return norm.sf(time, location, scale)


def do_fit(data, **kwargs) -> Tuple[float, float]:
    """Fits the provided data to the EXP distribution and estimates scale and location.

    :param data: the data to use in the fit.
    :return: (_location, _scale); the estimated parameters.
    :rtype: tuple
    """
    _location = kwargs.get("location", 0.0)  # Initial guess for location.
    _floc = kwargs.get("floc", None)  # Value to fix location parameter.
    _scale = kwargs.get("scale", 0.0)  # Initial guess for scale.
    _method = kwargs.get("method", "MLE")  # One of MLE or MM.

    if _floc is None:
        if scipy.__version__ >= "1.7.1":
            _location, _scale = norm.fit(
                data,
                loc=_location,
                scale=_scale,
                method=_method,
            )
        else:
            _location, _scale = norm.fit(
                data,
                loc=_location,
                scale=_scale,
            )
    else:
        if scipy.__version__ >= "1.7.1":
            _location, _scale = norm.fit(
                data,
                loc=_location,
                scale=_scale,
                floc=_floc,
                method=_method,
            )
        else:
            _location, _scale = norm.fit(
                data,
                loc=_location,
                scale=_scale,
                floc=_floc,
            )

    return _location, _scale



# -*- coding: utf-8 -*-
#
#       ramstk.models.commondb.database.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK program Database model."""

# Standard Library Imports
import gettext
from datetime import date, datetime, timedelta
from typing import Dict, Union

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKSiteConfiguration, RAMSTKUserConfiguration
from ramstk.db import BaseDatabase, do_create_program_db
from ramstk.models import RAMSTKSiteInfoRecord
from ramstk.models.commondb import (
    RAMSTKRPN,
    RAMSTKCategory,
    RAMSTKFailureMode,
    RAMSTKGroup,
    RAMSTKHazards,
    RAMSTKLoadHistory,
    RAMSTKManufacturer,
    RAMSTKMeasurement,
    RAMSTKMethod,
    RAMSTKModel,
    RAMSTKStakeholders,
    RAMSTKStatus,
    RAMSTKSubCategory,
    RAMSTKType,
    RAMSTKUser,
)

_ = gettext.gettext


class RAMSTKCommonDB:
    """The RAMSTK common database manager class.

    The RAMSTK common database manager is responsible for managing all the analysis,
    data, and matrix managers associated with the RAMSTK common database that
    is currently open.  The attributes of a RAMSTK common database manager are:

    :ivar dict dic_managers: a dict containing the instances of all the
        analysis, data, and matrix managers associated with the currently
        active run of RAMSTK.  The first key is the workstream module name
        which has a dict as a value.  The key of this second dict is the type
        of manager (analysis, data, matrix) and the value will be the instance
        of the applicable manager.
    :ivar program_dao: the BaseDatabase() object that will connect to the
        RAMSTK program database.
    :type program_dao: :class:`ramstk.db.base.BaseDatabase`
    """

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTK program manager."""
        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.
        self.dic_tables: Dict[str, object] = {
            "options": object,
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.site_configuration: RAMSTKSiteConfiguration = RAMSTKSiteConfiguration()
        self.common_dao: BaseDatabase = BaseDatabase()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_create_common, "request_create_common")
        # pub.subscribe(self.do_open_program, "request_open_program")
        # pub.subscribe(self.do_open_program, "succeed_create_program_database")
        # pub.subscribe(self.do_close_program, "request_close_program")
        # pub.subscribe(self.do_save_program, "request_update_program")

    def do_create_common(
        self,
        common_db: BaseDatabase,
        database: Dict[str, str],
        license_file: str,
    ) -> None:
        """Create a new RAMSTK Common database.

        :param common_db: the BaseDatabase() that is to be used to create and connect to
            the new RAMSTK common database.
        :param database: a dict containing the database connection arguments.
        :param license_file: the absolute path to the license file.
        :return: None
        :rtype: None
        """
        with open(
            self.site_configuration.RAMSTK_SITE_DIR
            + f'/{database["dialect"]}_common_db.sql',
            "r",
            encoding="utf-8",
        ) as _sql_file:
            self.common_dao = common_db
            do_create_program_db(database, _sql_file)
            self.common_dao.do_connect(database)
            self.do_load_site_info(license_file)

            _yn = (
                input(_("Would you like to add a RAMSTK Administrator? ([y]/n): "))
                or "y"
            )
            if _yn.lower() == "y":
                self.do_add_administrator()

            self.common_dao.do_disconnect()

            pub.sendMessage(
                "succeed_create_common_database",
                common_db=self.common_dao,
                database=database,
            )

    def do_add_administrator(self) -> None:
        """Add a new administrator to the RAMSTK pool."""
        _user = RAMSTKUser()
        _user.user_id = 0
        _user.user_group_id = 1

        _user.user_lname = input(  # nosec
            _("Enter the RAMSTK Administrator's last name (surname): ")
        )
        _user.user_fname = input(  # nosec
            _("Enter the RAMSTK Administrator's first name (given name): ")
        )
        _user.user_email = input(  # nosec
            _("Enter the RAMSTK Administrator's e-mail address: ")
        )
        _user.user_phone = input(  # nosec
            _("Enter the RAMSTK Administrator's phone number: ")
        )

        self.common_dao.session.add(_user)
        self.common_dao.session.commit()

    def do_load_site_info(self, license_file: str) -> None:
        """Load the Site Information table.

        :param license_file: the absolute path to the license file.
        :return: None
        :rtype: None
        """
        _dic_site_info: Dict[str, Union[date, int, str]] = {
            "site_name": "DEMO",
            "product_key": "DEMO",
            "expire_on": date.today() + timedelta(days=30),
            "function_enabled": 1,
            "requirement_enabled": 1,
            "hardware_enabled": 1,
            "software_enabled": 0,
            "rcm_enabled": 0,
            "testing_enabled": 0,
            "incident_enabled": 0,
            "survival_enabled": 0,
            "vandv_enabled": 1,
            "hazard_enabled": 1,
            "stakeholder_enabled": 1,
            "allocation_enabled": 1,
            "similar_item_enabled": 1,
            "fmea_enabled": 1,
            "pof_enabled": 1,
            "rbd_enabled": 0,
            "fta_enabled": 0,
        }
        _site_id: int = -1

        try:
            with open(
                license_file,
                "r",
                encoding="UTF-8",
            ) as _license_file:
                _contents = _license_file.readlines()
                _site_id = int(_contents[0].strip("\n"))
                _dic_site_info["product_key"] = _contents[1].strip("\n")
                _dic_site_info["expire_on"] = datetime.strptime(
                    _contents[2].strip("\n"), "%Y-%m-%d"
                )
                _dic_site_info["function_enabled"] = int(_contents[3].strip("\n"))
                _dic_site_info["requirement_enabled"] = int(_contents[4].strip("\n"))
                _dic_site_info["hardware_enabled"] = int(_contents[5].strip("\n"))
                _dic_site_info["software_enabled"] = int(_contents[6].strip("\n"))
                _dic_site_info["rcm_enabled"] = int(_contents[7].strip("\n"))
                _dic_site_info["testing_enabled"] = int(_contents[8].strip("\n"))
                _dic_site_info["incident_enabled"] = int(_contents[9].strip("\n"))
                _dic_site_info["survival_enabled"] = int(_contents[10].strip("\n"))
                _dic_site_info["vandv_enabled"] = int(_contents[11].strip("\n"))
                _dic_site_info["hazard_enabled"] = int(_contents[12].strip("\n"))
                _dic_site_info["stakeholder_enabled"] = int(_contents[13].strip("\n"))
                _dic_site_info["allocation_enabled"] = int(_contents[14].strip("\n"))
                _dic_site_info["similar_item_enabled"] = int(_contents[15].strip("\n"))
                _dic_site_info["fmea_enabled"] = int(_contents[16].strip("\n"))
                _dic_site_info["pof_enabled"] = int(_contents[17].strip("\n"))
                _dic_site_info["rbd_enabled"] = int(_contents[18].strip("\n"))
                _dic_site_info["fta_enabled"] = int(_contents[19].strip("\n"))
                _dic_site_info["site_name"] = _contents[20].strip("\n")
        except IOError:
            _error_msg = (
                "Unable to read license key file.  Defaulting to a 30-day demo license."
            )
            pub.sendMessage("fail_read_license", error_message=_error_msg)

        _site_info = RAMSTKSiteInfoRecord()
        _site_info.site_id = _site_id
        _site_info.set_attributes(_dic_site_info)

        self.common_dao.session.add(_site_info)
        self.common_dao.session.commit()

    def do_load_site_variables(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTKUserConfiguration global variables from the site db.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        pub.sendMessage(
            "do_log_info_msg",
            logger_name="INFO",
            message="Loading global RAMSTK configuration variables.",
        )

        user_configuration = self._do_load_action_variables(user_configuration)
        user_configuration = self._do_load_hardware_variables(user_configuration)
        user_configuration = self._do_load_incident_variables(user_configuration)
        user_configuration = self._do_load_miscellaneous_variables(user_configuration)
        user_configuration = self._do_load_pof_variables(user_configuration)
        user_configuration = self._do_load_requirement_variables(user_configuration)
        user_configuration = self._do_load_rpn_variables(user_configuration)
        user_configuration = self._do_load_severity(user_configuration)
        user_configuration = self._do_load_user_workgroups(user_configuration)

        pub.sendMessage(
            "do_log_info_msg",
            logger_name="INFO",
            message="Loaded global RAMSTK configuration variables.",
        )

        return user_configuration

    def _do_load_action_variables(
        self,
        user_configuration: RAMSTKUserConfiguration,
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTK_ACTION_CATEGORY variable.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKCategory)
            .filter(RAMSTKCategory.category_type == "action")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_ACTION_CATEGORY[_record.category_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["category_type"],
                _attributes["value"],
            )
        for _record in (
            self.common_dao.session.query(RAMSTKStatus)
            .filter(RAMSTKStatus.status_type == "action")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_ACTION_STATUS[_record.status_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["status_type"],
            )

        return user_configuration

    def _do_load_hardware_variables(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load variables associated with hardware categories and failure modes.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :param site_db: the RAMSTK Site Database to read the values of the
            global variables.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKCategory)
            .filter(RAMSTKCategory.category_type == "hardware")
            .all()
        ):

            _subcats = {}
            user_configuration.RAMSTK_FAILURE_MODES[_record.category_id] = {}
            user_configuration.RAMSTK_STRESS_LIMITS[_record.category_id] = (
                _record.harsh_ir_limit,
                _record.mild_ir_limit,
                _record.harsh_pr_limit,
                _record.mild_pr_limit,
                _record.harsh_vr_limit,
                _record.mild_vr_limit,
                _record.harsh_deltat_limit,
                _record.mild_deltat_limit,
                _record.harsh_maxt_limit,
                _record.mild_maxt_limit,
            )
            for _subcat in (
                self.common_dao.session.query(RAMSTKSubCategory)
                .filter(RAMSTKSubCategory.category_id == _record.category_id)
                .all()
            ):
                _subcats[_subcat.subcategory_id] = _subcat.description

                _modes = {}
                user_configuration.RAMSTK_FAILURE_MODES[_record.category_id][
                    _subcat.subcategory_id
                ] = {}
                for _mode in (
                    self.common_dao.session.query(RAMSTKFailureMode)
                    .filter(RAMSTKFailureMode.category_id == _record.category_id)
                    .filter(RAMSTKFailureMode.subcategory_id == _subcat.subcategory_id)
                    .all()
                ):
                    _modes[_mode.mode_id] = [
                        _mode.description,
                        _mode.mode_ratio,
                        _mode.source,
                    ]

                user_configuration.RAMSTK_FAILURE_MODES[_record.category_id][
                    _subcat.subcategory_id
                ] = _modes

            user_configuration.RAMSTK_CATEGORIES[
                _record.category_id
            ] = _record.description
            user_configuration.RAMSTK_SUBCATEGORIES[_record.category_id] = _subcats

        return user_configuration

    def _do_load_incident_variables(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTK_INCIDENT_CATEGORY variable.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKCategory)
            .filter(RAMSTKCategory.category_type == "incident")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_INCIDENT_CATEGORY[_record.category_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["category_type"],
                _attributes["value"],
            )
        for _record in (
            self.common_dao.session.query(RAMSTKStatus)
            .filter(RAMSTKStatus.status_type == "incident")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_INCIDENT_STATUS[_record.status_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["status_type"],
            )
        for _record in (
            self.common_dao.session.query(RAMSTKType)
            .filter(RAMSTKType.type_type == "incident")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_INCIDENT_TYPE[_record.type_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["type_type"],
            )

        return user_configuration

    def _do_load_miscellaneous_variables(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load miscellaneous variables that don't fit in another grouping.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKMethod)
            .filter(RAMSTKMethod.method_type == "detection")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_DETECTION_METHODS[_record.method_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["method_type"],
            )
        for _record in self.common_dao.session.query(RAMSTKHazards).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_HAZARDS[_record.hazard_id] = (
                _attributes["hazard_category"],
                _attributes["hazard_subcategory"],
            )
        for _record in self.common_dao.session.query(RAMSTKManufacturer).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_MANUFACTURERS[_record.manufacturer_id] = (
                _attributes["description"],
                _attributes["location"],
                _attributes["cage_code"],
            )
        for _record in (
            self.common_dao.session.query(RAMSTKMeasurement)
            .filter(RAMSTKMeasurement.measurement_type == "unit")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_MEASUREMENT_UNITS[_record.measurement_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["measurement_type"],
            )
        for _record in (
            self.common_dao.session.query(RAMSTKType)
            .filter(RAMSTKType.type_type == "validation")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_VALIDATION_TYPE[_record.type_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["type_type"],
            )

        return user_configuration

    def _do_load_pof_variables(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTK_DAMAGE_MODELS variable.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKModel)
            .filter(RAMSTKModel.model_type == "damage")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_DAMAGE_MODELS[_record.model_id] = _attributes[
                "description"
            ]
        for _record in self.common_dao.session.query(RAMSTKLoadHistory).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_LOAD_HISTORY[_record.history_id] = _attributes[
                "description"
            ]
        for _record in (
            self.common_dao.session.query(RAMSTKMeasurement)
            .filter(RAMSTKMeasurement.measurement_type == "damage")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_MEASURABLE_PARAMETERS[_record.measurement_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["measurement_type"],
            )

        return user_configuration

    def _do_load_requirement_variables(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load variables related to requiremetents and stakeholders.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKGroup)
            .filter(RAMSTKGroup.group_type == "affinity")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_AFFINITY_GROUPS[_record.group_id] = (
                _attributes["description"],
                _attributes["group_type"],
            )
        for _record in (
            self.common_dao.session.query(RAMSTKType)
            .filter(RAMSTKType.type_type == "requirement")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_REQUIREMENT_TYPE[_record.type_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["type_type"],
            )
        for _record in self.common_dao.session.query(RAMSTKStakeholders).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_STAKEHOLDERS[
                _record.stakeholders_id
            ] = _attributes["stakeholder"]

        return user_configuration

    def _do_load_rpn_variables(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RPN detection, occurremce, and severity variables.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKRPN)
            .filter(RAMSTKRPN.rpn_type == "detection")
            .all()
        ):
            user_configuration.RAMSTK_RPN_DETECTION[
                _record.value
            ] = _record.get_attributes()

        for _record in (
            self.common_dao.session.query(RAMSTKRPN)
            .filter(RAMSTKRPN.rpn_type == "occurrence")
            .all()
        ):
            user_configuration.RAMSTK_RPN_OCCURRENCE[
                _record.value
            ] = _record.get_attributes()

        for _record in (
            self.common_dao.session.query(RAMSTKRPN)
            .filter(RAMSTKRPN.rpn_type == "severity")
            .all()
        ):
            user_configuration.RAMSTK_RPN_SEVERITY[
                _record.value
            ] = _record.get_attributes()

        return user_configuration

    def _do_load_severity(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTK_SEVERITY variable.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKCategory)
            .filter(RAMSTKCategory.category_type == "risk")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_SEVERITY[_record.category_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["category_type"],
                _attributes["value"],
            )

        return user_configuration

    def _do_load_user_workgroups(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTK_USERS and RAMSTK_WORKGROUPS variables.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in self.common_dao.session.query(RAMSTKUser).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_USERS[_record.user_id] = (
                _attributes["user_lname"],
                _attributes["user_fname"],
                _attributes["user_email"],
                _attributes["user_phone"],
                _attributes["user_group_id"],
            )
        for _record in (
            self.common_dao.session.query(RAMSTKGroup)
            .filter(RAMSTKGroup.group_type == "workgroup")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_WORKGROUPS[_record.group_id] = (
                _attributes["description"],
                _attributes["group_type"],
            )

        return user_configuration



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.design_electric.components.connection.py is part of the RAMSTK
#       Project.
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Connection Input Panel."""

# Standard Library Imports
from typing import Any, Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import _
from ramstk.views.gtk3.widgets import RAMSTKComboBox, RAMSTKEntry, RAMSTKFixedPanel


class ConnectionDesignElectricInputPanel(RAMSTKFixedPanel):
    """Displays connection assessment input attribute data.

    The Connection assessment input view displays all the assessment inputs for
    the selected connection.  This includes, currently, inputs for
    MIL-HDBK-217FN2 parts count and part stress analysis.  The attributes of a
    Connection assessment input view are:

    :cvar dict _dic_quality: dictionary of MIL-HDBK-217 capacitor quality
        levels.  Key is capacitor subcategory ID; values are lists of quality
        levels.
    :cvar dict _dic_insert: dictionary of connector insert materials.  First
        key is connector type ID, second key is connector specification ID;
        values are lists of insert materials.
     :cvar dict _dic_insert: dictionary of connector insert materials.  First
        key is connector type ID, second key is connector specification ID;
        values are lists of insert materials.
    :cvar dict _dic_type: dictionary of connector types.  Key is connector
        subcategory ID; values are lists of types.

    :ivar list _lst_labels: list of label text to display for the capacitor
        MIL-HDBK-217 input parameters.
    :ivar _lst_widgets: the list of widgets to display in the panel.  These
        are listed in the order they should appear on the panel.

    :ivar _hazard_rate_method_id: the ID of the method to use for estimating
        the Hardware item's hazard rate.
    :ivar _subcategory_id: the ID of the Hardware item's subcategory.
    :ivar _title: the text to put on the RAMSTKFrame() holding the
        assessment input widgets.

    :ivar fmt: the formatting to use when displaying float values.
    :ivar cmbInsert: select and display the available insert materials for the
        connector.
    :ivar cmbSpecification: select and display the governing specification of
        the connection.
    :ivar cmbType: select and display the type of the connection.

    :ivar txtActivePins: enter and display the number of active pins in the
        connector.
    :ivar txtAmpsContact: enter and display the amps carried by the pins in the
        connector.
    :ivar txtContactGauge: enter and display the contact gauge of the
        connector.
    :ivar txtMating: enter and display the number of mate/demate cycles the
        connector undergoes per 1000 hours.
    :ivar txtNHand: enter and display the number of hand soldered PTH
        connections.
    :ivar txtNPlanes: enter and display the number of layers in the circuit
        board the PTH needs to penetrate.
    :ivar txtNWave: enter and display the number of wave soldered PTH
        connections.
    """

    # Define private dict class attributes.

    # Quality levels; key is the subcategory ID.
    _dic_quality: Dict[int, List[List[str]]] = {
        1: [["MIL-SPEC"], [_("Lower")]],
        2: [["MIL-SPEC"], [_("Lower")]],
        4: [[_("MIL-SPEC or comparable IPC standards")], [_("Lower")]],
        5: [
            [_("Automated")],
            [_("Manual, Upper")],
            [_("Manual, Standard")],
            [_("Manual, Lower")],
        ],
    }

    # Connector types; key is the subcategory ID.
    _dic_type: Dict[int, List[List[str]]] = {
        1: [
            [_("Rack and Panel")],
            [_("Circular")],
            [_("Power")],
            [_("Coaxial")],
            [_("Triaxial")],
        ],
        4: [
            [_("PWA/PCB with PTHs")],
            [
                _(
                    "Discrete Wiring with Electroless Deposited PTH (<3 Levels "
                    "of Circuitry)"
                )
            ],
        ],  # noqa
        5: [
            [_("Hand Solder w/o Wrapping")],
            [_("Hand Solder w/ Wrapping")],
            [_("Crimp")],
            [_("Weld")],
            [_("Solderless Wrap")],
            [_("Clip Termination")],
            [_("Reflow Solder")],
        ],
    }

    # Specifications; key is the type ID.
    _dic_specification: Dict[int, List[List[str]]] = {
        1: [
            [_("MIL-C-24308")],
            [_("MIL-C-28748")],
            [_("MIL-C-28804")],
            [_("MIL-C-83513")],
            [_("MIL-C-83733")],
        ],
        2: [
            [_("MIL-C-5015")],
            [_("MIL-C-26482")],
            [_("MIL-C-28840")],
            [_("MIL-C-38999")],
            [_("MIL-C-81511")],
            [_("MIL-C-83723")],
        ],
        3: [[_("MIL-C-3767")], [_("MIL-C-22992")]],
        4: [
            [_("MIL-C-3607")],
            [_("MIL-C-3643")],
            [_("MIL-C-3650")],
            [_("MIL-C-3655")],
            [_("MIL-C-25516")],
            [_("MIL-C-39012")],
            [_("MIL-C-55235")],
            [_("MIL-C-55339")],
        ],
        5: [[_("MIL-C-49142")]],
    }

    _lst_insert_A: List[List[str]] = [
        [_("Vitreous Glass")],
        [_("Alumina Ceramic")],
        [_("Polyimide")],
    ]
    _lst_insert_B: List[List[str]] = [
        [_("Diallylphtalate")],
        [_("Melamine")],
        [_("Flourosilicone")],
        [_("Silicone Rubber")],
        [_("Polysulfone")],
        [_("Epoxy Resin")],
    ]
    _lst_insert_C: List[List[str]] = [
        [_("Polytetraflourethylene (Teflon)")],
        [_("Chlorotriflourethylene (Kel-f)")],
    ]
    _lst_insert_D: List[List[str]] = [
        [_("Polyamide (Nylon)")],
        [_("Polychloroprene (Neoprene)")],
        [_("Polyethylene")],
    ]
    # Connector insert material; first key is the type ID, second key is the
    # specification ID.
    _dic_insert: Dict[int, Dict[int, List[List[str]]]] = {
        1: {
            1: _lst_insert_B,
            2: _lst_insert_B,
            3: _lst_insert_A + _lst_insert_B,
            4: _lst_insert_A + _lst_insert_B,
            5: _lst_insert_A + _lst_insert_B,
        },
        2: {
            1: _lst_insert_B + _lst_insert_D,
            2: _lst_insert_A + _lst_insert_B + _lst_insert_D,
            3: _lst_insert_A + _lst_insert_B,
            4: _lst_insert_A + _lst_insert_B,
            5: _lst_insert_B,
            6: _lst_insert_B,
        },
        3: {1: _lst_insert_B + _lst_insert_D, 2: _lst_insert_B + _lst_insert_D},
        4: {
            1: _lst_insert_C,
            2: _lst_insert_C,
            3: _lst_insert_C,
            4: _lst_insert_C,
            5: _lst_insert_C,
            6: _lst_insert_C,
            7: _lst_insert_C,
            8: _lst_insert_B + _lst_insert_C,
        },
        5: {1: _lst_insert_B + _lst_insert_C},
    }

    # Define private list attributes.

    # Define private scalar class attributes.
    _record_field: str = "hardware_id"
    _select_msg: str = "selected_hardware"
    _tag: str = "design_electric"
    _title: str = _("Connection Design Inputs")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Connection assessment input view."""
        super().__init__()

        # Initialize widgets.
        self.cmbInsert: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbQuality: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbSpecification: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbType: RAMSTKComboBox = RAMSTKComboBox()
        self.txtContactGauge: RAMSTKEntry = RAMSTKEntry()
        self.txtActivePins: RAMSTKEntry = RAMSTKEntry()
        self.txtAmpsContact: RAMSTKEntry = RAMSTKEntry()
        self.txtMating: RAMSTKEntry = RAMSTKEntry()
        self.txtNWave: RAMSTKEntry = RAMSTKEntry()
        self.txtNHand: RAMSTKEntry = RAMSTKEntry()
        self.txtNPlanes: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._hazard_rate_method_id: int = 0
        self._quality_id: int = 0

        # Initialize public dictionary attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "quality_id": [
                32,
                self.cmbQuality,
                "changed",
                super().on_changed_combo,
                "wvw_editing_reliability",
                0,
                {
                    "tooltip": _("The quality level of the connector/connection."),
                },
                _("Quality Level:"),
                "gint",
            ],
            "type_id": [
                48,
                self.cmbType,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The type of connector/connection."),
                },
                _("Connector Type:"),
                "gint",
            ],
            "specification_id": [
                36,
                self.cmbSpecification,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The governing specification for the connection."),
                },
                _("Specification:"),
                "gint",
            ],
            "insert_id": [
                18,
                self.cmbInsert,
                "changed",
                super().on_changed_combo,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The connector insert material."),
                },
                _("Insert Material:"),
                "gint",
            ],
            "contact_gauge": [
                8,
                self.txtContactGauge,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                22,
                {
                    "tooltip": _("The gauge of the contacts in the connector."),
                },
                _("Contact Gauge:"),
                "gint",
            ],
            "n_active_pins": [
                22,
                self.txtActivePins,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The number of active pins in the connector."),
                },
                _("Active Pins:"),
                "gint",
            ],
            "current_operating": [
                10,
                self.txtAmpsContact,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0.0,
                {
                    "tooltip": _("The amperes per active contact."),
                },
                _("Amperes/Contact:"),
                "gfloat",
            ],
            "n_cycles": [
                24,
                self.txtMating,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _(
                        "The number of connector mate and unmate cycles per 1000 hours "
                        "of operation."
                    ),
                },
                _("Mating/Unmating Cycles (per 1000 hours):"),
                "gfloat",
            ],
            "n_wave_soldered": [
                27,
                self.txtNWave,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The number of wave soldered PTH connections."),
                },
                _("Number of Wave Soldered PTH:"),
                "gint",
            ],
            "n_hand_soldered": [
                26,
                self.txtNHand,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _("The number of hand soldered PTH connections."),
                },
                _("Number of Hand Soldered PTH:"),
                "gint",
            ],
            "n_circuit_planes": [
                23,
                self.txtNPlanes,
                "changed",
                super().on_changed_entry,
                f"wvw_editing_{self._tag}",
                0,
                {
                    "tooltip": _(
                        "The number of circuit planes for wave soldered connections."
                    ),
                },
                _("Number of Circuit Planes:"),
                "gint",
            ],
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.category_id: int = 0
        self.subcategory_id: int = 0

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.cmbSpecification.connect("changed", self._do_load_insert)
        self.cmbType.connect("changed", self._do_load_specification)

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self.do_load_comboboxes,
            "changed_subcategory",
        )
        pub.subscribe(
            self._do_set_reliability_attributes,
            "succeed_get_reliability_attributes",
        )

    def do_load_comboboxes(self, subcategory_id: int) -> None:
        """Load the connection RKTComboBox()s.

        :param subcategory_id: the subcategory ID of the selected connection.
        :return: None
        :rtype: None
        """
        self.subcategory_id = subcategory_id

        # Load the quality level RAMSTKComboBox().
        if self._hazard_rate_method_id == 1:  # MIL-HDBK-217F parts count.
            _data = [["MIL-SPEC"], [_("Lower")]]
        else:
            try:
                _data = self._dic_quality[self.subcategory_id]
            except KeyError:
                _data = []
        self.cmbQuality.do_load_combo(_data, signal="changed")

        # Load the connector type RAMSTKComboBox().
        try:
            _data = self._dic_type[self.subcategory_id]
        except KeyError:
            _data = []
        self.cmbType.do_load_combo(_data, signal="changed")

        # Clear the remaining ComboBox()s.  These are loaded dynamically
        # based on the selection made in other ComboBox()s.
        _model = self.cmbSpecification.get_model()
        _model.clear()

        _model = self.cmbInsert.get_model()
        _model.clear()

    def _do_load_insert(self, combo: RAMSTKComboBox) -> None:
        """Load the insert RAMSTKComboBox() when the specification changes.

        :param combo: the specification RAMSTKCombo() that called this method.
        :return: None
        :rtype: None
        """
        try:
            _type_id = int(self.cmbType.get_active())
            _spec_id = int(combo.get_active())
            _inserts = self._dic_insert[_type_id][_spec_id]
        except KeyError:
            _inserts = []
        self.cmbInsert.do_load_combo(entries=_inserts, signal="changed")

    def _do_load_specification(self, combo: RAMSTKComboBox) -> None:
        """Retrieve RAMSTKCombo() changes and assign to Connection attribute.

        :param combo: the connection type RAMSTKCombo() that called this
            method.
        :return: None
        :rtype: None
        """
        try:
            _type_id = int(combo.get_active())
            _specifications = self._dic_specification[_type_id]
        except KeyError:
            _specifications = []
        self.cmbSpecification.do_load_combo(entries=_specifications, signal="changed")

    def _do_set_reliability_attributes(self, attributes: Dict[str, Any]) -> None:
        """Set the attributes when the reliability attributes are retrieved.

        :param attributes: the dict of reliability attributes.
        :return: None
        :rtype: None
        """
        if attributes["hardware_id"] == self._record_id:
            self._hazard_rate_method_id = attributes["hazard_rate_method_id"]
            self._quality_id = attributes["quality_id"]

            self.cmbQuality.set_sensitive(True)
            self.cmbQuality.do_update(
                self._quality_id,
                signal="changed",
            )

            pub.sendMessage(
                f"request_get_{self._tag}_attributes",
                node_id=self._record_id,
            )

    def _do_set_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set widget sensitivity as needed for the selected connection.

        :return: None
        :rtype: None
        """
        self.cmbInsert.set_sensitive(False)
        self.cmbSpecification.set_sensitive(False)
        self.cmbType.set_sensitive(False)

        self.txtActivePins.set_sensitive(False)
        self.txtAmpsContact.set_sensitive(False)
        self.txtContactGauge.set_sensitive(False)
        self.txtMating.set_sensitive(False)
        self.txtNHand.set_sensitive(False)
        self.txtNPlanes.set_sensitive(False)
        self.txtNWave.set_sensitive(False)

        if self._hazard_rate_method_id == 1:
            self.cmbType.set_sensitive(True)
            self.cmbType.do_update(
                attributes["type_id"],
                signal="changed",
            )
        else:
            self.__do_set_circular_sensitive(attributes)
            self.__do_set_ic_socket_sensitive(attributes)
            self.__do_set_pwa_edge_sensitive(attributes)
            self.__do_set_pth_sensitive(attributes)

    def __do_set_circular_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the widgets for circular connectors sensitive or not.

        :return: None
        :rtype: None
        """
        if self.subcategory_id == 1:
            self.cmbType.set_sensitive(True)
            # We don't block the callback signal otherwise the specification
            # RAMSTKComboBox() will not be loaded and set.
            self.cmbType.set_active(attributes["type_id"])

            self.cmbSpecification.set_sensitive(True)
            # We don't block the callback signal otherwise the insert
            # RAMSTKComboBox() will not be loaded and set.
            self.cmbSpecification.set_active(attributes["specification_id"])

            self.cmbInsert.set_sensitive(True)
            self.cmbInsert.do_update(
                attributes["insert_id"],
                signal="changed",
            )

            self.txtActivePins.set_sensitive(True)
            self.txtActivePins.do_update(
                str(attributes["n_active_pins"]),
                signal="changed",
            )

            self.txtAmpsContact.set_sensitive(True)
            self.txtAmpsContact.do_update(
                str(self.fmt.format(attributes["current_operating"])),
                signal="changed",
            )

            self.txtContactGauge.set_sensitive(True)
            self.txtContactGauge.do_update(
                str(attributes["contact_gauge"]),
                signal="changed",
            )

            self.txtMating.set_sensitive(True)
            self.txtMating.do_update(
                str(attributes["n_cycles"]),
                signal="changed",
            )

    def __do_set_ic_socket_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the widgets for IC socket connectors sensitive or not.

        :return: None
        :rtype: None
        """
        if self.subcategory_id == 3:
            self.cmbQuality.set_sensitive(False)
            self.txtActivePins.set_sensitive(True)
            self.txtActivePins.do_update(
                str(attributes["n_active_pins"]),
                signal="changed",
            )

    def __do_set_pwa_edge_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the widgets for PCB/PWA edge connectors sensitive or not.

        :return: None
        :rtype: None
        """
        if self.subcategory_id == 2:
            self.txtAmpsContact.set_sensitive(True)
            self.txtAmpsContact.do_update(
                str(self.fmt.format(attributes["current_operating"])),
                signal="changed",
            )

            self.txtContactGauge.set_sensitive(True)
            self.txtContactGauge.do_update(
                str(attributes["contact_gauge"]),
                signal="changed",
            )

            self.txtMating.set_sensitive(True)
            self.txtMating.do_update(
                str(attributes["n_cycles"]),
                signal="changed",
            )

            self.txtActivePins.set_sensitive(True)
            self.txtActivePins.do_update(
                str(attributes["n_active_pins"]),
                signal="changed",
            )

    def __do_set_pth_sensitive(self, attributes: Dict[str, Any]) -> None:
        """Set the widgets for PTH connections sensitive or not.

        :return: None
        :rtype: None
        """
        if self.subcategory_id == 4:
            self.txtNWave.set_sensitive(True)
            self.txtNWave.do_update(
                str(attributes["n_wave_soldered"]),
                signal="changed",
            )
            self.txtNHand.set_sensitive(True)
            self.txtNHand.do_update(
                str(attributes["n_hand_soldered"]),
                signal="changed",
            )
            self.txtNPlanes.set_sensitive(True)
            self.txtNPlanes.do_update(
                str(attributes["n_circuit_planes"]),
                signal="changed",
            )



# pylint: disable=unused-import, missing-docstring
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.options.panel.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK GTK3 Options Panels."""

# Standard Library Imports
from typing import Any, Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import _
from ramstk.views.gtk3.widgets import RAMSTKCheckButton, RAMSTKEntry, RAMSTKFixedPanel


class OptionsPanel(RAMSTKFixedPanel):
    """The panel to display options to be edited."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_get_siteinfo_attributes"
    _tag = "option"
    _title = _("General Information")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Edit Options panel."""
        super().__init__()

        # Initialize widgets.
        self.chkFunctions: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Function Module Enabled")
        )
        self.chkRequirements: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Requirements Module Enabled")
        )
        self.chkHardware: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Hardware Module Enabled")
        )
        self.chkValidation: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Validation Module Enabled")
        )
        self.chkHazards: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Hazards Analysis Module Enabled")
        )
        self.chkStakeholder: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Stakeholder Analysis Module Enabled")
        )
        self.chkAllocation: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("R(t) Allocation Module Enabled")
        )
        self.chkSimilarItem: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Similar Item Analysis Module Enabled")
        )
        self.chkFMEA: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("(D)FME(C)A Module Enabled")
        )
        self.chkPoF: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Physics of Failure (PoF) Module Enabled")
        )

        self.txtSiteID: RAMSTKEntry = RAMSTKEntry()
        self.txtSiteName: RAMSTKEntry = RAMSTKEntry()
        self.txtProductKey: RAMSTKEntry = RAMSTKEntry()
        self.txtExpireDate: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "site_id": [
                0,
                self.txtSiteID,
                "changed",
                super().on_changed_entry,
                "",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Site ID:"),
                "gint",
            ],
            "site_name": [
                1,
                self.txtSiteName,
                "changed",
                super().on_changed_entry,
                "",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Site Name:"),
                "gchararray",
            ],
            "product_key": [
                2,
                self.txtProductKey,
                "changed",
                super().on_changed_entry,
                "",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Product Key:"),
                "gchararray",
            ],
            "expire_on": [
                3,
                self.txtExpireDate,
                "changed",
                super().on_changed_entry,
                "",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Expire Date:"),
                "gchararray",
            ],
            "function_enabled": [
                4,
                self.chkFunctions,
                "toggled",
                super().on_toggled,
                "",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _(""),
                "gint",
            ],
            "requirement_enabled": [
                5,
                self.chkRequirements,
                "toggled",
                super().on_toggled,
                "",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _(""),
                "gint",
            ],
            "hardware_enabled": [
                6,
                self.chkHardware,
                "toggled",
                super().on_toggled,
                "",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _(""),
                "gint",
            ],
            "validation_enabled": [
                7,
                self.chkValidation,
                "toggled",
                super().on_toggled,
                "",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _(""),
                "gint",
            ],
            "hazard_enabled": [
                8,
                self.chkHazards,
                "toggled",
                super().on_toggled,
                "",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _(""),
                "gint",
            ],
            "stakeholder_enabled": [
                9,
                self.chkStakeholder,
                "toggled",
                super().on_toggled,
                "",
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _(""),
                "gint",
            ],
            "allocation_enabled": [
                10,
                self.chkAllocation,
                "toggled",
                super().on_toggled,
                "",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _(""),
                "gint",
            ],
            "similar_item_enabled": [
                11,
                self.chkSimilarItem,
                "toggled",
                super().on_toggled,
                "",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _(""),
                "gint",
            ],
            "fmea_enabled": [
                12,
                self.chkFMEA,
                "toggled",
                super().on_toggled,
                "",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _(""),
                "gint",
            ],
            "pof_enabled": [
                13,
                self.chkPoF,
                "toggled",
                super().on_toggled,
                "",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _(""),
                "gint",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.

        pub.sendMessage("request_get_option_attributes2", attributes={"site_id": 1})



# -*- coding: utf-8 -*-
#
#       ramstk.analyses.SimilarItem.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Reliability Similar Item Assessment Module."""

# Standard Library Imports
from typing import Dict, List, Tuple

# Third Party Imports
# noinspection PyPackageRequirements
from sympy import symbols, sympify  # type: ignore

ENVIRONMENT_FROM_TO: Dict[Tuple[int, int], float] = {
    (0, 0): 1.0,
    (1, 1): 1.0,
    (1, 2): 0.2,
    (1, 3): 0.3,
    (1, 4): 0.3,
    (1, 5): 0.1,
    (1, 6): 1.1,
    (2, 1): 5.0,
    (2, 2): 1.0,
    (2, 3): 1.4,
    (2, 4): 1.4,
    (2, 5): 0.5,
    (2, 6): 5.0,
    (3, 1): 3.3,
    (3, 2): 0.7,
    (3, 3): 1.0,
    (3, 4): 1.0,
    (3, 5): 0.3,
    (3, 6): 3.3,
    (4, 1): 3.3,
    (4, 2): 0.7,
    (4, 3): 1.0,
    (4, 4): 1.0,
    (4, 5): 0.3,
    (4, 6): 3.3,
    (5, 1): 10.0,
    (5, 2): 2.0,
    (5, 3): 3.3,
    (5, 4): 3.3,
    (5, 5): 1.0,
    (5, 6): 10.0,
    (6, 1): 0.9,
    (6, 2): 0.2,
    (6, 3): 0.3,
    (6, 4): 0.3,
    (6, 5): 0.1,
    (6, 6): 1.0,
}
QUALITY_FROM_TO: Dict[Tuple[int, int], float] = {
    (0, 0): 1.0,
    (1, 1): 1.0,
    (1, 2): 0.8,
    (1, 3): 0.5,
    (1, 4): 0.2,
    (2, 1): 1.3,
    (2, 2): 1.0,
    (2, 3): 0.6,
    (2, 4): 0.3,
    (3, 1): 2.0,
    (3, 2): 1.7,
    (3, 3): 1.0,
    (3, 4): 0.4,
    (4, 1): 5.0,
    (4, 2): 3.3,
    (4, 3): 2.5,
    (4, 4): 1.0,
}
TEMPERATURE_FROM_TO: Dict[Tuple[float, float], float] = {
    (10.0, 10.0): 1.0,
    (10.0, 20.0): 0.9,
    (10.0, 30.0): 0.8,
    (10.0, 40.0): 0.8,
    (10.0, 50.0): 0.7,
    (10.0, 60.0): 0.5,
    (10.0, 70.0): 0.4,
    (20.0, 10.0): 1.1,
    (20.0, 20.0): 1.0,
    (20.0, 30.0): 0.9,
    (20.0, 40.0): 0.8,
    (20.0, 50.0): 0.7,
    (20.0, 60.0): 0.6,
    (20.0, 70.0): 0.5,
    (30.0, 10.0): 1.2,
    (30.0, 20.0): 1.1,
    (30.0, 30.0): 1.0,
    (30.0, 40.0): 0.9,
    (30.0, 50.0): 0.8,
    (30.0, 60.0): 0.6,
    (30.0, 70.0): 0.5,
    (40.0, 10.0): 1.3,
    (40.0, 20.0): 1.2,
    (40.0, 30.0): 1.1,
    (40.0, 40.0): 1.0,
    (40.0, 50.0): 0.9,
    (40.0, 60.0): 0.7,
    (40.0, 70.0): 0.6,
    (50.0, 10.0): 1.5,
    (50.0, 20.0): 1.4,
    (50.0, 30.0): 1.2,
    (50.0, 40.0): 1.1,
    (50.0, 50.0): 1.0,
    (50.0, 60.0): 0.8,
    (50.0, 70.0): 0.7,
    (60.0, 10.0): 1.9,
    (60.0, 20.0): 1.7,
    (60.0, 30.0): 1.6,
    (60.0, 40.0): 1.5,
    (60.0, 50.0): 1.2,
    (60.0, 60.0): 1.0,
    (60.0, 70.0): 0.8,
    (70.0, 10.0): 2.4,
    (70.0, 20.0): 2.2,
    (70.0, 30.0): 1.9,
    (70.0, 40.0): 1.8,
    (70.0, 50.0): 1.5,
    (70.0, 60.0): 1.2,
    (70.0, 70.0): 1.0,
}


# noinspection PyTypeChecker
def calculate_topic_633(
    environment: Dict[str, float],
    quality: Dict[str, float],
    temperature: Dict[str, float],
    hazard_rate: float,
) -> Tuple[float, float, float, float]:
    """Calculate the Similar Item analysis using Topic 6.3.3 approach.

    This method calculates the new hazard rate using the approach found
    in The Reliability Toolkit: Commercial Practices Edition, Topic 6.3.3.

    :param environment: the active environment ID for the from and to
        environments.
    :param quality: the quality level ID for the from and to quality.
    :param temperature: the ambient operating temperature (in C) for the
        from and to temperatures.
    :param hazard_rate: the current hazard rate of the hardware item
        being calculated.
    :return: (_change_factor_1, _change_factor_2, _change_factor_3, _result_1);
        the three change factors (quality, environment, and temperature) and
        the assessment result.
    :rtype: tuple
    :raise: KeyError if passed an environment, quality, or temperature dict
        that is missing the from, to, or both keys.
    :raise: TypeError if passed a string value for either temperature.
    """
    # Convert user-supplied temperatures to whole values used in Topic 633.
    temperature["from"] = round(temperature["from"] / 10.0) * 10.0
    temperature["to"] = round(temperature["to"] / 10.0) * 10.0

    _change_factor_1 = QUALITY_FROM_TO[
        (quality["from"], quality["to"])  # type: ignore
    ]  # type: ignore
    _change_factor_2 = ENVIRONMENT_FROM_TO[
        (environment["from"], environment["to"])  # type: ignore
    ]  # type: ignore
    _change_factor_3 = TEMPERATURE_FROM_TO[(temperature["from"], temperature["to"])]

    _result_1 = float(
        hazard_rate / (_change_factor_1 * _change_factor_2 * _change_factor_3)
    )

    return _change_factor_1, _change_factor_2, _change_factor_3, _result_1


# pylint: disable=too-many-locals
def calculate_user_defined(sia: Dict[str, float]):
    """Calculate the user-defined similar item analysis.

    :param sia: the user-defined similar item assessment dict.  The
        calling method/function should create the sia dict as follows:

        sia = OrderedDict({
            _key: None
            for _key in [
                'hr', 'pi1', 'pi2', 'pi3', 'pi3', 'pi4', 'pi5', 'pi6', 'pi7',
                'pi8', 'pi9', 'pi10', 'uf1', 'uf2', 'uf3', 'uf4', 'uf5', 'ui1',
                'ui2', 'ui3', 'ui4', 'ui5', 'equation1', 'equation2',
                'equation3', 'equation4', 'equation5', 'res1', 'res2', 'res3',
                'res4', 'res5'
                ]
            })

    :return: sia; the similar item assessment dict with updated results.
    :rtype: dict
    """
    (
        hr,
        pi1,
        pi2,
        pi3,
        pi4,
        pi5,
        pi6,
        pi7,
        pi8,
        pi9,
        pi10,
        uf1,
        uf2,
        uf3,
        uf4,
        uf5,
        ui1,
        ui2,
        ui3,
        ui4,
        ui5,
        res1,
        res2,
        res3,
        res4,
        res5,
    ) = symbols(
        "hr pi1 pi2 pi3 pi4 pi5 pi6 pi7 pi8 pi9 pi10 uf1 uf2 uf3 uf4 uf5 ui1 "
        "ui2 ui3 ui4 ui5 res1 res2 res3 res4 res5"
    )

    # The subs argument needs to be passed as a dict of sia values just like
    # it is below.  This will result in duplicate code warnings, but passing
    # it like this is required to allow the use of the results in subsequent
    # calculations.
    # pylint: disable=eval-used
    sia["res1"] = sympify(sia["equation1"]).evalf(
        subs={
            hr: sia["hr"],
            pi1: sia["pi1"],
            pi2: sia["pi2"],
            pi3: sia["pi3"],
            pi4: sia["pi4"],
            pi5: sia["pi5"],
            pi6: sia["pi6"],
            pi7: sia["pi7"],
            pi8: sia["pi8"],
            pi9: sia["pi9"],
            pi10: sia["pi10"],
            uf1: sia["uf1"],
            uf2: sia["uf2"],
            uf3: sia["uf3"],
            uf4: sia["uf4"],
            uf5: sia["uf5"],
            ui1: sia["ui1"],
            ui2: sia["ui2"],
            ui3: sia["ui3"],
            ui4: sia["ui4"],
            ui5: sia["ui5"],
            res1: sia["res1"],
            res2: sia["res2"],
            res3: sia["res3"],
            res4: sia["res4"],
            res5: sia["res5"],
        }
    )
    sia["res2"] = sympify(sia["equation2"]).evalf(
        subs={
            hr: sia["hr"],
            pi1: sia["pi1"],
            pi2: sia["pi2"],
            pi3: sia["pi3"],
            pi4: sia["pi4"],
            pi5: sia["pi5"],
            pi6: sia["pi6"],
            pi7: sia["pi7"],
            pi8: sia["pi8"],
            pi9: sia["pi9"],
            pi10: sia["pi10"],
            uf1: sia["uf1"],
            uf2: sia["uf2"],
            uf3: sia["uf3"],
            uf4: sia["uf4"],
            uf5: sia["uf5"],
            ui1: sia["ui1"],
            ui2: sia["ui2"],
            ui3: sia["ui3"],
            ui4: sia["ui4"],
            ui5: sia["ui5"],
            res1: sia["res1"],
            res2: sia["res2"],
            res3: sia["res3"],
            res4: sia["res4"],
            res5: sia["res5"],
        }
    )
    sia["res3"] = sympify(sia["equation3"]).evalf(
        subs={
            hr: sia["hr"],
            pi1: sia["pi1"],
            pi2: sia["pi2"],
            pi3: sia["pi3"],
            pi4: sia["pi4"],
            pi5: sia["pi5"],
            pi6: sia["pi6"],
            pi7: sia["pi7"],
            pi8: sia["pi8"],
            pi9: sia["pi9"],
            pi10: sia["pi10"],
            uf1: sia["uf1"],
            uf2: sia["uf2"],
            uf3: sia["uf3"],
            uf4: sia["uf4"],
            uf5: sia["uf5"],
            ui1: sia["ui1"],
            ui2: sia["ui2"],
            ui3: sia["ui3"],
            ui4: sia["ui4"],
            ui5: sia["ui5"],
            res1: sia["res1"],
            res2: sia["res2"],
            res3: sia["res3"],
            res4: sia["res4"],
            res5: sia["res5"],
        }
    )
    sia["res4"] = sympify(sia["equation4"]).evalf(
        subs={
            hr: sia["hr"],
            pi1: sia["pi1"],
            pi2: sia["pi2"],
            pi3: sia["pi3"],
            pi4: sia["pi4"],
            pi5: sia["pi5"],
            pi6: sia["pi6"],
            pi7: sia["pi7"],
            pi8: sia["pi8"],
            pi9: sia["pi9"],
            pi10: sia["pi10"],
            uf1: sia["uf1"],
            uf2: sia["uf2"],
            uf3: sia["uf3"],
            uf4: sia["uf4"],
            uf5: sia["uf5"],
            ui1: sia["ui1"],
            ui2: sia["ui2"],
            ui3: sia["ui3"],
            ui4: sia["ui4"],
            ui5: sia["ui5"],
            res1: sia["res1"],
            res2: sia["res2"],
            res3: sia["res3"],
            res4: sia["res4"],
            res5: sia["res5"],
        }
    )
    sia["res5"] = sympify(sia["equation5"]).evalf(
        subs={
            hr: sia["hr"],
            pi1: sia["pi1"],
            pi2: sia["pi2"],
            pi3: sia["pi3"],
            pi4: sia["pi4"],
            pi5: sia["pi5"],
            pi6: sia["pi6"],
            pi7: sia["pi7"],
            pi8: sia["pi8"],
            pi9: sia["pi9"],
            pi10: sia["pi10"],
            uf1: sia["uf1"],
            uf2: sia["uf2"],
            uf3: sia["uf3"],
            uf4: sia["uf4"],
            uf5: sia["uf5"],
            ui1: sia["ui1"],
            ui2: sia["ui2"],
            ui3: sia["ui3"],
            ui4: sia["ui4"],
            ui5: sia["ui5"],
            res1: sia["res1"],
            res2: sia["res2"],
            res3: sia["res3"],
            res4: sia["res4"],
            res5: sia["res5"],
        }
    )

    return sia


def set_user_defined_change_factors(
    sia: Dict[str, float], factors: List[float]
) -> Dict[str, float]:
    """Set the change factors for the user-defined calculations.

    :param sia: the similar item assessment dict.
    :param list factors: the list of change factors; the list items are float
        or int.
    :return: sia; the similar item assessment dict with updated factor values.
    :rtype: dict
    """
    # Get the change factor values.
    for _idx in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]:
        _key = list(sia.keys())[_idx]
        try:
            sia[_key] = factors[_idx - 1]
        except IndexError:
            sia[_key] = 0.0

    return sia


def set_user_defined_floats(
    sia: Dict[str, float], floats: List[float]
) -> Dict[str, float]:
    """Set the user-defined float values for the user-defined calculations.

    :param sia: the similar item assessment dict.
    :param list floats: the list of float values.
    :return: sia; the similar item assessment dict with updated float values.
    :rtype: dict
    """
    for _idx in [11, 12, 13, 14, 15]:
        _key = list(sia.keys())[_idx]
        try:
            sia[_key] = float(floats[_idx - 11])
        except IndexError:
            sia[_key] = 0.0

    return sia


def set_user_defined_ints(sia: Dict[str, int], ints: List[int]) -> Dict[str, int]:
    """Set the user-defined integer values for the user-defined calculations.

    :param sia: the similar item assessment dict.
    :param list ints: the list of integer values.
    :return: sia; the similar item assessment dict with updated integer values.
    :rtype: dict
    """
    for _idx in [16, 17, 18, 19, 20]:
        _key = list(sia.keys())[_idx]
        try:
            sia[_key] = int(ints[_idx - 16])
        except IndexError:
            sia[_key] = 0

    return sia


def set_user_defined_functions(
    sia: Dict[str, str], functions: List[str]
) -> Dict[str, str]:
    """Set the user-defined functions for the user-defined calculations.

    :param sia: the similar item assessment dict.
    :param list functions: the list of functions; list items are str.
    :return: sia; the similar item assessment dict with updated functions.
    :rtype: dict
    """
    for _idx in [21, 22, 23, 24, 25]:
        _key = list(sia.keys())[_idx]
        try:
            sia[_key] = str(functions[_idx - 21])
        except IndexError:
            sia[_key] = ""

    return sia


def set_user_defined_results(
    sia: Dict[str, float], results: List[float]
) -> Dict[str, float]:
    """Set the user-defined results for the user-defined calculations.

    This allows the use of the results fields to be manually set to float
    values by the user essentially creating five more user-defined float
    values.

    :param sia: the similar item assessment dict.
    :param list results: the list of results.
    :return: sia; the similar item assessment dict with updated results.
    :rtype: dict
    """
    for _idx in [26, 27, 28, 29, 30]:
        _key = list(sia.keys())[_idx]
        try:
            sia[_key] = results[_idx - 26]
        except IndexError:
            sia[_key] = 0.0

    return sia



# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKDialog as RAMSTKDialog
from ramstk.views.gtk3.widgets import RAMSTKPanel as RAMSTKPanel

# RAMSTK Local Imports
from . import OptionsPanel as OptionsPanel

class OptionsDialog(RAMSTKDialog):
    _pnlPanel: OptionsPanel
    def __init__(self, parent: object = ...) -> None: ...
    def _cancel(self, __button: Gtk.Button): ...
    def __make_ui(self) -> None: ...



# Standard Library Imports
from typing import Any, Dict, List

# RAMSTK Package Imports
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKComboBox as RAMSTKComboBox
from ramstk.views.gtk3.widgets import RAMSTKEntry as RAMSTKEntry
from ramstk.views.gtk3.widgets import RAMSTKFixedPanel as RAMSTKFixedPanel

class CapacitorDesignElectricInputPanel(RAMSTKFixedPanel):
    _dic_quality: Dict[int, List[Any]]
    _dic_specifications: Dict[int, List[Any]]
    _dic_styles: Dict[int, List[Any]]
    _record_field: str
    _select_msg: str
    _tag: str
    _title: str
    cmbConfiguration: RAMSTKComboBox
    cmbConstruction: RAMSTKComboBox
    cmbQuality: RAMSTKComboBox
    cmbSpecification: RAMSTKComboBox
    cmbStyle: RAMSTKComboBox
    txtCapacitance: RAMSTKEntry
    txtESR: RAMSTKEntry
    _hazard_rate_method_id: int
    _quality_id: int
    dic_attribute_widget_map: Dict[str, List[Any]]
    category_id: int
    subcategory_id: int
    def __init__(self) -> None: ...
    def do_load_comboboxes(self, subcategory_id: int) -> None: ...
    def _do_set_reliability_attributes(self, attributes: Dict[str, Any]) -> None: ...
    def _do_load_styles(self, combo: RAMSTKComboBox) -> None: ...
    def _do_set_sensitive(self, attributes: Dict[str, Any]) -> None: ...
    def __do_set_parts_count_sensitive(self, attributes: Dict[str, Any]) -> None: ...
    def __do_set_part_stress_sensitive(self, attributes: Dict[str, Any]) -> None: ...



# Standard Library Imports
from typing import Any

# RAMSTK Package Imports
from ramstk.analyses.statistics import (
    do_calculate_beta_bounds as do_calculate_beta_bounds,
)
from ramstk.db import RAMSTK_BASE as RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord as RAMSTKBaseRecord

class RAMSTKValidationRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Any
    __tablename__: str
    __table_args__: Any
    revision_id: Any
    validation_id: Any
    acceptable_maximum: Any
    acceptable_mean: Any
    acceptable_minimum: Any
    acceptable_variance: Any
    confidence: Any
    cost_average: Any
    cost_ll: Any
    cost_maximum: Any
    cost_mean: Any
    cost_minimum: Any
    cost_ul: Any
    cost_variance: Any
    date_end: Any
    date_start: Any
    description: Any
    measurement_unit: Any
    name: Any
    status: Any
    task_specification: Any
    task_type: Any
    time_average: Any
    time_ll: Any
    time_maximum: Any
    time_mean: Any
    time_minimum: Any
    time_ul: Any
    time_variance: Any
    def get_attributes(self): ...
    def calculate_task_time(self) -> None: ...
    def calculate_task_cost(self) -> None: ...



# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE as RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord as RAMSTKBaseRecord

class RAMSTKEnvironmentRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Dict[str, Any]
    __tablename__: str
    __table_args__: Dict[str, bool]
    revision_id: int
    mission_id: int
    phase_id: int
    environment_id: int
    name: str
    units: str
    minimum: float
    maximum: float
    mean: float
    variance: float
    ramp_rate: float
    low_dwell_time: float
    high_dwell_time: float
    phase: relationship
    is_mission: bool
    is_phase: bool
    is_env: bool
    def get_attributes(self) -> Dict[str, Any]: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.milhdbk217f.panel.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 MIL-HDBK-217F Panels."""

# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3.widgets import RAMSTKEntry, RAMSTKFixedPanel, RAMSTKLabel


class MilHdbk217FResultPanel(RAMSTKFixedPanel):
    """Display Hardware assessment results attribute data.

    The widgets of a MIL-HDBK-217F result panel are:

    :ivar txtLambdaB: displays the base hazard rate of the hardware item.
    :ivar txtPiQ: displays the quality factor for the hardware item.
    :ivar txtPiE: displays the environment factor for the hardware item.

    The attributes of a MIL-HDBK-217F result panel are:

    :ivar _hazard_rate_method_id: the ID of the method selected to calculate the
        hazard rate of the selected component.
    :ivar _lambda_b: the base hazard rate of the selected component.
    :ivar category_id: the hardware category ID of the selected component.
    :ivar subcategory_id: the hardware subcategory ID of the selected component.
    """

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field: str = "hardware_id"
    _select_msg: str = "selected_hardware"
    _tag: str = "milhdbk217f"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Hardware assessment result view."""
        super().__init__()

        # Initialize widgets.
        self.lblModel: RAMSTKLabel = RAMSTKLabel("")
        self.txtLambdaB: RAMSTKEntry = RAMSTKEntry()
        self.txtPiQ: RAMSTKEntry = RAMSTKEntry()
        self.txtPiE: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._hazard_rate_method_id: int = 0
        self._lambda_b: float = 0.0

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.category_id: int = 0
        self.subcategory_id: int = 0

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self._do_set_hardware_attributes,
            "succeed_get_hardware_attributes",
        )
        pub.subscribe(
            self._do_set_reliability_attributes,
            "succeed_get_reliability_attributes",
        )

    def do_load_entries(self, attributes: Dict[str, Any]) -> None:
        """Load the Hardware assessment results page.

        :param attributes: the attribute dict for the selected Hardware.
        :return: None
        :rtype: None
        """
        self.txtLambdaB.set_sensitive(False)
        self.txtPiE.set_sensitive(False)
        self.txtPiQ.set_sensitive(False)

        # Display the correct calculation model.
        self.__do_set_model_label()

        self.txtLambdaB.do_update(
            str(self.fmt.format(self._lambda_b or 0.0)),
            signal="changed",
        )
        self.txtPiQ.do_update(
            str(self.fmt.format(attributes["piQ"] or 1.0)),
            signal="changed",
        )
        self.txtPiE.do_update(
            str(self.fmt.format(attributes["piE"] or 1.0)),
            signal="changed",
        )

    def _do_set_hardware_attributes(self, attributes: Dict[str, Any]) -> None:
        """Set the attributes when the reliability attributes are retrieved.

        :param attributes: the dict of reliability attributes.
        :return: None
        :rtype: None
        """
        if attributes["hardware_id"] == self._record_id:
            self.category_id = attributes["category_id"]
            self.subcategory_id = attributes["subcategory_id"]

    def _do_set_reliability_attributes(self, attributes: Dict[str, Any]) -> None:
        """Set the attributes when the reliability attributes are retrieved.

        :param attributes: the dict of reliability attributes.
        :return: None
        :rtype: None
        """
        if attributes["hardware_id"] == self._record_id:
            self._hazard_rate_method_id = attributes["hazard_rate_method_id"]
            self._lambda_b = attributes["lambda_b"]

            pub.sendMessage(
                f"request_get_{self._tag}_attributes",
                node_id=self._record_id,
            )

    def __do_set_model_label(self) -> None:
        """Set the text displayed in the hazard rate model RAMSTKLabel().

        :return: None
        :rtype: None
        """
        if self._hazard_rate_method_id == 1:
            self.lblModel.set_markup(
                '<span foreground="blue">\u03BB<sub>p</sub> = '
                "\u03BB<sub>b</sub>\u03C0<sub>Q</sub></span> "
            )
        elif self._hazard_rate_method_id == 2:
            try:
                self.lblModel.set_markup(self._dic_part_stress[self.subcategory_id])
            except KeyError:
                self.lblModel.set_markup("No Model")
        else:
            self.lblModel.set_markup("No Model")



# pylint: disable=non-parent-init-called
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.widgets.treeview.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKTreeView Module."""

# Standard Library Imports
import datetime
from typing import Any, Callable, Dict, List, Tuple

# Third Party Imports
import toml
import treelib

# RAMSTK Package Imports
from ramstk.utilities import deprecated, string_to_boolean
from ramstk.views.gtk3 import Gdk, GdkPixbuf, GObject, Gtk, Pango

# RAMSTK Local Imports
from .label import RAMSTKLabel
from .widget import RAMSTKWidget


def do_make_column(cells: List[object], **kwargs: Dict[str, Any]) -> Gtk.TreeViewColumn:
    """Make a Gtk.TreeViewColumn().

    :param list cells: list of Gtk.CellRenderer()s that are to be packed in
        the column.
    :return: _column
    :rtype: :class:`Gtk.TreeViewColumn`
    """
    _heading = kwargs.get("heading", "")
    _visible = kwargs.get("visible", True)

    _column = Gtk.TreeViewColumn("")

    for _cell in cells:
        if isinstance(_cell, Gtk.CellRendererPixbuf):
            _column.pack_start(_cell, False)
        else:
            _column.pack_start(_cell, True)

    _label = RAMSTKLabel(_heading)  # type: ignore
    _label.do_set_properties(width=-1, height=-1, justify=Gtk.Justification.CENTER)
    _column.set_widget(_label)
    _column.set_resizable(True)
    _column.set_alignment(0.5)
    _column.set_visible(_visible)

    return _column


def do_set_cell_properties(cell: object, properties: Dict[str, Any]) -> None:
    """Set common properties of Gtk.CellRenderers().

    :param cell: the cell whose properties are to be set.
    :param properties: the properties for the cell.
    :return: None
    :rtype: None
    """
    _bg_color = properties.get("bg_color", "#FFFFFF")
    _digits = properties.get("digits", 2)
    _editable = properties.get("editable", False)
    _fg_color = properties.get("fg_color", "#000000")
    _lower = properties.get("lower", 1)
    _step = properties.get("step", 1)
    _upper = properties.get("upper", 10)
    _visible = properties.get("visible", True)
    _weight = properties.get("weight", 400)
    _weight_set = properties.get("weight_set", False)

    if not _editable:
        _color = Gdk.RGBA(255.0, 255.0, 255.0, 1.0)
        _fg_color = "#000000"
        cell.set_property("cell-background-rgba", _color)  # type: ignore

    cell.set_property("visible", _visible)  # type: ignore
    cell.set_property("yalign", 0.1)  # type: ignore

    if isinstance(cell, Gtk.CellRendererCombo):
        _cellmodel = Gtk.ListStore(GObject.TYPE_STRING)
        _cellmodel.append([""])
        cell.set_property("editable", _editable)
        cell.set_property("has-entry", False)
        cell.set_property("model", _cellmodel)
        cell.set_property("text-column", 0)
    elif isinstance(cell, Gtk.CellRendererSpin):
        _adjustment = Gtk.Adjustment(lower=_lower, upper=_upper, step_incr=_step)
        cell.set_property("adjustment", _adjustment)
        cell.set_property("digits", _digits)
        cell.set_property("editable", _editable)
    elif isinstance(cell, Gtk.CellRendererText):
        cell.set_property("background", _bg_color)
        cell.set_property("editable", _editable)
        cell.set_property("foreground", _fg_color)
        cell.set_property("weight", _weight)
        cell.set_property("weight-set", _weight_set)
        cell.set_property("wrap-width", 250)
        cell.set_property("wrap-mode", Pango.WrapMode.WORD)
    elif isinstance(cell, Gtk.CellRendererToggle):
        cell.set_property("activatable", _editable)
        cell.set_property("cell-background", _bg_color)


class RAMSTKTreeView(Gtk.TreeView, RAMSTKWidget):
    """The RAMSTKTreeView class."""

    def __init__(self) -> None:
        """Initialize a RAMSTKTreeView() instance.

        :return: None
        :rtype: None
        """
        # noinspection PyCallByClass,PyTypeChecker
        RAMSTKWidget.__init__(self)
        GObject.GObject.__init__(self)

        # Initialize private dictionary instance attributes:
        self.dic_row_loader: Dict[str, Callable] = {}

        # Initialize private list instance attributes:

        # Initialize private scalar instance attributes.
        self._has_pixbuf: bool = False

        # Initialize public dictionary instance attributes.
        self.cellprops: Dict[str, Any] = {}
        self.datatypes: Dict[str, str] = {}
        self.editable: Dict[str, bool] = {}
        self.headings: Dict[str, str] = {}
        self.position: Dict[str, int] = {}
        self.visible: Dict[str, bool] = {}
        self.widgets: Dict[str, object] = {}

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.
        self.filt_model: Gtk.TreeModelFilter = Gtk.TreeModelFilter()
        self.selection = self.get_selection()
        self.unfilt_model = self.get_model()

    def do_change_cell(
        self,
        cell: Gtk.CellRenderer,
        path: str,
        new_row: Gtk.TreeIter,
        position: int,
    ) -> Any:
        """Handle Gtk.CellRenderer() edits.

        :param cell: the Gtk.CellRenderer() that was edited.
        :param path: the Gtk.TreeView() path of the Gtk.CellRenderer() that
            was edited.
        :param new_row: the new Gtk.TreeIter() selected in the
            Gtk.CellRendererCombo().  This is relative to the cell renderer's model,
            not the RAMSTKTreeView() model.
        :param position: the column position of the edited
            Gtk.CellRenderer().
        :return: new_text; the value of the new text converted to the
            correct data type for the attribute being edited.
        """
        _model = self.get_model()
        _cell_model = cell.get_property("model")

        _new_text = _cell_model.get_value(new_row, 0)

        _idx = 0
        _iter = _cell_model.get_iter_first()
        while _iter is not None:
            if _cell_model.get_value(_iter, 0) == _new_text:
                _model[path][position] = _idx
                break
            _iter = _cell_model.iter_next(_iter)
            _idx += 1

        return _idx

    def do_edit_cell(
        self, cell: Gtk.CellRenderer, path: str, new_text: Any, position: int
    ) -> Any:
        """Handle Gtk.CellRenderer() edits.

        :param cell: the Gtk.CellRenderer() that was edited.
        :param path: the Gtk.TreeView() path of the Gtk.CellRenderer() that
            was edited.
        :param new_text: the new text in the edited Gtk.CellRenderer().
        :param position: the column position of the edited
            Gtk.CellRenderer().
        :return: new_text; the value of the new text converted to the
            correct data type for the attribute being edited.
        :rtype: Any
        """
        _model = self.get_model()
        _convert = GObject.type_name(_model.get_column_type(position))

        if isinstance(cell, Gtk.CellRendererToggle):
            new_text = not cell.get_active()
        elif _convert == "gchararray":
            new_text = str(new_text)
        elif _convert == "gint":
            try:
                new_text = int(new_text)
            except ValueError:
                new_text = int(float(new_text))
        elif _convert == "gfloat":
            new_text = float(new_text)

        _model[path][position] = new_text

        return new_text

    @deprecated
    def do_expand_tree(self) -> None:
        """Expand the RAMSTKTreeView().

        Currently unused.  Determine if it has value.

        :return: None
        :rtype: None
        """
        _model = self.get_model()
        try:
            _row = _model.get_iter_first()
        except AttributeError:
            _row = None

        self.expand_all()
        if _row is not None:
            _path = _model.get_path(_row)
            _column = self.get_column(0)
            self.set_cursor(_path, None, False)
            self.row_activated(_path, _column)

    @deprecated
    def do_get_row_by_value(self, search_col: int, value: Any) -> Gtk.TreeIter:
        """Find the row in the RAMSTKTreeView() containing the passed value.

        Currently unused.  Determine if it has value.

        :param search_col: the column number to search for the desired value.
        :param value: the value to match.
        :return: _iter; the Gtk.TreeIter() for the matching row.
        :rtype: :class:`Gtk.TreeIter`
        """
        _model = self.get_model()
        try:
            _row = _model.get_iter_first()
        except AttributeError:
            _row = None

        while _row is not None:
            _value = _model.get_value(_row, search_col)
            # pylint: disable=no-else-return
            if _value == value:
                return _row
            else:
                _row = _model.iter_next(_row)

    def do_insert_row(self, data: Dict[str, Any], prow: Gtk.TreeIter = None) -> None:
        """Insert a new row in the treeview.

        :param data: the data dictionary for the new row to insert.
        :param prow: the parent row of the row to insert.
        :return: None
        :rtype: None
        """
        _data = []
        _model, _row = self.selection.get_selected()

        for _key in self.position:
            _data.append(data[_key])

        _row = _model.append(prow, _data)

        _path = _model.get_path(_row)
        _column = self.get_column(0)
        self.set_cursor(_path, None, False)
        self.row_activated(_path, _column)

    def do_load_combo_cell(self, index: int, items: List[str]) -> None:
        """Load items into cell with combobox.

        :param index: the index in the RAMSTKTreeView() model of the
            Gtk.CellRendererCombo() to load.
        :param items: the list of entries to load into the Gtk.CellRendererCombo().
        :return: None
        :rtype: None
        """
        _model = self.get_cell_model(index)
        for _item in items:
            _model.append([_item])

    def do_load_tree(self, tree: treelib.Tree, row: Gtk.TreeIter = None) -> None:
        """Load the RAMSTKTreeView with the contents of the tree."""
        _row = None
        _node = tree.get_node(tree.root)

        if _node.data is not None:
            _row = self.dic_row_loader[_node.tag](_node, row)

        for _n in tree.children(_node.identifier):
            self.do_load_tree(tree.subtree(_n.identifier), _row)

    def do_make_columns(self) -> None:
        """Make the columns for the RAMSTKTreeView().

        :return: None
        :rtype: None
        """
        for _key, _position in self.position.items():
            _cell = self.widgets[_key]
            _properties = {"editable": self.editable[_key], **self.cellprops[_key]}
            do_set_cell_properties(
                _cell,
                properties=_properties,
            )
            # If creating a RAMSTKTreeView() that displays icons and this is
            # the first column we're creating, add a Gtk.CellRendererPixbuf()
            # to go along with the data in the first column.
            if self._has_pixbuf and _key == "col0":
                _pbcell = Gtk.CellRendererPixbuf()
                _pbcell.set_property("xalign", 0.5)
                _pbcell.set_property("cell-background", _properties["bg_color"])
                _column = do_make_column(
                    [_pbcell, _cell],
                    heading="",  # type: ignore
                    visible=True,  # type: ignore
                )
                _column.set_attributes(_pbcell, pixbuf=len(self.position.values()))
            else:
                _column = do_make_column(
                    [_cell],
                    heading=self.headings[_key],  # type: ignore
                    visible=string_to_boolean(self.visible[_key]),  # type: ignore
                )
            _column.set_cell_data_func(
                _cell, self._do_format_cell, (_position, self.datatypes[_key])
            )

            self._do_set_column_properties(_key, _column)
            self.append_column(_column)

    def do_make_model(self) -> None:
        """Make the RAMSTKTreeView() data model.

        :return: None
        :rtype: None
        """
        _types = []

        # Create a list of GObject data types to pass to the model.
        for __, _datatype in self.datatypes.items():  # pylint: disable=unused-variable
            _types.append(GObject.type_from_name(_datatype))

        if self._has_pixbuf:
            _types.append(GdkPixbuf.Pixbuf)

        self.unfilt_model = Gtk.TreeStore(*_types)
        self.set_model(self.unfilt_model)

    # noinspection PyTypeChecker
    def do_parse_format(self, fmt_file: str) -> None:
        """Parse the format file for the RAMSTKTreeView().

        :param fmt_file: the absolute path to the format file to read.
        :return: None
        :rtype: None
        """
        _format = toml.load(fmt_file)

        self._has_pixbuf = string_to_boolean(_format["pixbuf"])

        self.position = {}
        _keys = sorted(_format["position"], key=_format["position"].get)
        for _key in _keys:
            self.position[_key] = _format["position"][_key]
            self.editable[_key] = self.editable[_key] or string_to_boolean(
                _format["editable"][_key]
            )
            self.visible[_key] = self.visible[_key] or string_to_boolean(
                _format["visible"][_key]
            )

        self.headings = _format["usertitle"]

    def do_set_editable_columns(self, method: object) -> None:
        """Set the treeview columns editable or read-only.

        :param method: the callback method for the cell.
        :return: None
        """
        for _key in self.editable:
            _column = self.get_column(self.position[_key])

            # Get the last cell so those columns containing a
            # Gtk.CellRendererPixBuf() will return the Gtk.CellRendererText()
            # that is packed with it.  If there is only one cell renderer,
            # that is the one that will be returned.
            _cell = _column.get_cells()[-1]

            if isinstance(self.widgets[_key], Gtk.CellRendererToggle):
                _cell.connect("toggled", method, None, self.position[_key])
            elif isinstance(
                self.widgets[_key],
                (Gtk.CellRendererSpin, Gtk.CellRendererText),
            ):
                _cell.connect("edited", method, self.position[_key])

    def do_set_visible_columns(self) -> None:
        """Set the treeview columns visible or hidden.

        :return: None
        :rtype: None
        """
        for _key, _visible in self.visible.items():
            _column = self.get_column(self.position[_key])
            _column.set_visible(_visible)

    def get_aggregate_attributes(self, entity: object) -> List[Any]:
        """Get the attributes for aggregate work stream modules.

        :param entity: the RAMSTK Program database table whose attributes
            are to be returned.
        :return: _attributes; a list of the attributes values in the order
            they will be displayed.
        :rtype: list
        """
        _attributes = []
        try:
            for _key in self.position:
                if _key == "dict":
                    _attributes.append(str(entity))
                else:
                    try:
                        if isinstance(entity[_key], datetime.date):  # type: ignore
                            entity[_key] = entity[_key].strftime(  # type: ignore
                                "%Y-%m-%d"
                            )
                        entity[_key] = entity[_key].decode("utf-8")  # type: ignore
                    except (AttributeError, KeyError):
                        pass
                    _attributes.append(entity[_key])  # type: ignore
        except TypeError:
            pass

        return _attributes

    def get_cell_model(self, column: int, clear: bool = True) -> Gtk.TreeModel:
        """Retrieve the Gtk.TreeModel() from a Gtk.CellRendererCombo().

        :param column: the column number to retrieve the cell's model.
        :param clear: whether or not to clear the Gtk.TreeModel().
            Default is True.
        :return: _model
        :rtype: :class:`Gtk.TreeModel`
        """
        _column = self.get_column(column)
        _cell = _column.get_cells()[0]
        _model = _cell.get_property("model")

        if clear:
            _model.clear()

        return _model

    def get_simple_attributes(self, entity: object) -> List[Any]:
        """Get the attributes for simple work stream modules.

        :param entity: the RAMSTK Program database table whose attributes
            are to be returned.
        :return: _attributes; a list of the attributes values in the order
            they will be displayed.
        :rtype: list
        """
        _attributes = []
        _temp = entity.get_attributes()  # type: ignore

        for _key in self.position:
            try:
                if isinstance(_temp[_key], datetime.date):
                    _temp[_key] = _temp[_key].strftime("%Y-%m-%d")
                _temp[_key] = _temp[_key].decode("utf-8")
            except (AttributeError, KeyError):
                pass
            _attributes.append(_temp[_key])

        return _attributes

    @staticmethod
    def _do_format_cell(
        __column: Gtk.TreeViewColumn,
        cell: Gtk.CellRenderer,
        model: Gtk.TreeModel,
        row: Gtk.TreeIter,
        data: Tuple[Any],
    ) -> None:
        """Set the formatting of the Gtk.Treeview() Gtk.CellRenderers().

        :param __column: the Gtk.TreeViewColumn() containing the
            Gtk.CellRenderer() to format.
        :type __column: :class:`Gtk.TreeViewColumn`
        :param cell: the Gtk.CellRenderer() to format.
        :type cell: :class:`Gtk.CellRenderer`
        :param model: the Gtk.TreeModel() containing the Gtk.TreeViewColumn().
        :type model: :class:`Gtk.TreeModel`
        :param row: the Gtk.TreeIter() pointing to the row containing the
            Gtk.CellRenderer() to format.
        :type row: :class:`Gtk.TreeIter`
        :param tuple data: a tuple containing the position and the data type.
        """
        if data[1] == "gfloat":  # type: ignore
            fmt = "{0:0.6g}"
        elif data[1] == "gint":  # type: ignore
            fmt = "{0:0d}"
        else:
            return

        val = model.get_value(row, data[0])
        try:
            cell.set_property("text", fmt.format(val))
        except TypeError:  # It's a Gtk.CellRendererToggle
            pass
        except ValueError:
            pass

    def _do_set_column_properties(self, key: str, column: Gtk.TreeViewColumn) -> None:
        """Set the properties of the RAMSTKTreeView() column.

        :param key: the value of the key in the widgets and position dicts.
        :param column: the Gtk.TreeViewColumn() to set properties.
        :return: None
        :rtype: None
        """
        _cell = column.get_cells()[-1]

        if isinstance(self.widgets[key], Gtk.CellRendererToggle):
            column.set_attributes(_cell, active=self.position[key])
        elif isinstance(
            self.widgets[key],
            (
                Gtk.CellRendererCombo,
                Gtk.CellRendererSpin,
                Gtk.CellRendererText,
            ),
        ):
            column.set_attributes(_cell, text=self.position[key])

        if self.position[key] > 0:
            column.set_reorderable(True)

    @staticmethod
    def _resize_wrap(
        column: Gtk.TreeViewColumn, __param, cell: Gtk.CellRenderer
    ) -> None:
        """Dynamically set wrap-width property for a Gtk.CellRenderer().

        This is called whenever the column width in the Gtk.TreeView() is
        resized.

        :param column: the Gtk.TreeViewColumn() being resized.
        :param GParamInt __param: the triggering parameter.
        :param cell: the Gtk.CellRenderer() that needs to be resized.
        :return: None
        :rtype: None
        """
        _width = column.get_width()

        if _width <= 0:
            _width = 25
        else:
            _width += 10

            try:
                cell.set_property("wrap-width", _width)
            except TypeError:  # This is a Gtk.CellRendererToggle
                cell.set_property("width", _width)


class CellRendererML(Gtk.CellRendererText):
    """Create a multi-line cell renderer."""

    def __init__(self) -> None:
        """Initialize a CellRendererML instance."""
        GObject.GObject.__init__(self)

        self.textedit_window = None
        self.selection = None
        self.treestore = None
        self.treeiter = None

        self.textedit = Gtk.TextView()
        self.textbuffer = self.textedit.get_buffer()

    # pylint: disable=arguments-differ
    def do_get_size(self, widget, cell_area):
        """Get the size of the CellRendererML.

        :param widget:
        :param cell_area:
        """
        size_tuple = Gtk.CellRendererText.do_get_size(self, widget, cell_area)

        return size_tuple

    # pylint: disable=arguments-differ,too-many-locals
    def do_start_editing(
        self, __event, treeview, path, __background_area, cell_area, __flags
    ):
        """Handle edits of the CellRendererML.

        :param __event:
        :param treeview:
        :param path:
        :param __background_area:
        :param cell_area:
        :param __flags:
        """
        if not self.get_property("editable"):
            return

        self.selection = treeview.get_selection()
        self.treestore, self.treeiter = self.selection.get_selected()

        self.textedit_window = Gtk.Dialog(parent=treeview.get_toplevel())
        self.textedit_window.action_area.hide()
        self.textedit_window.set_decorated(False)
        self.textedit_window.set_property("skip-taskbar-hint", True)
        self.textedit_window.set_transient_for(None)

        self.textedit.set_editable(True)
        self.textedit.set_property("visible", True)
        self.textbuffer.set_property("text", self.get_property("text"))

        self.textedit_window.connect("key-press-event", self._keyhandler)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(
            Gtk.PolicyType.AUTOMATIC,
            Gtk.PolicyType.AUTOMATIC,
        )
        scrolled_window.set_property("visible", True)
        # self.textedit_window.vbox.pack_start(scrolled_window, True, True, 0)

        scrolled_window.add(self.textedit)
        self.textedit_window.vbox.add(scrolled_window)
        self.textedit_window.realize()

        # Position the popup below the edited cell (and try hard to keep the
        # popup within the toplevel window)
        (tree_x, tree_y) = treeview.get_bin_window().get_origin()
        (tree_w, tree_h) = treeview.window.get_geometry()[2:4]
        (t_w, t_h) = self.textedit_window.window.get_geometry()[2:4]
        x_pos = tree_x + min(
            cell_area.x,
            tree_w - t_w + treeview.get_visible_rect().x,
        )
        y_pos = tree_y + min(
            cell_area.y,
            tree_h - t_h + treeview.get_visible_rect().y,
        )
        self.textedit_window.move(x_pos, y_pos)
        self.textedit_window.resize(cell_area.width, cell_area.height)

        # Run the dialog, get response by tracking keypresses
        response = self.textedit_window.run()

        if response == Gtk.ResponseType.OK:
            self.textedit_window.destroy()

            (iter_first, iter_last) = self.textbuffer.get_bounds()
            text = self.textbuffer.get_text(iter_first, iter_last)

            treeview.set_cursor(path, None, False)

            self.emit("edited", path, text)

        elif response == Gtk.ResponseType.CANCEL:
            self.textedit_window.destroy()
        else:
            print(f"response {response} received")
            self.textedit_window.destroy()

    def _keyhandler(self, __widget, event):
        """Handle key-press-events on the Gtk.TextView().

        :param __widget: the Gtk.TextView() that called this method.
        :param event: the Gdk.Event() that called this method.
        """
        _keyname = Gdk.keyval_name(event.keyval)

        if (
            event.get_state()
            & (Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.CONTROL_MASK)
            and _keyname == "Return"
        ):
            self.textedit_window.response(Gtk.ResponseType.OK)


# Register the new widget types.
GObject.type_register(RAMSTKTreeView)
GObject.type_register(CellRendererML)



# pylint: disable=non-parent-init-called
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.widgets.frame.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK GTK3 frame Module."""

# Standard Library Imports
from typing import Any, Dict, Union

# RAMSTK Package Imports
from ramstk.views.gtk3 import GObject, Gtk

# RAMSTK Local Imports
from .label import RAMSTKLabel


class RAMSTKFrame(Gtk.Frame):
    """The RAMSTK Frame class."""

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTK Frame."""
        GObject.GObject.__init__(self)

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set the RAMSTKFrame properties."""
        _bold: Union[Dict[str, Any], bool] = kwargs.get("bold", False)
        _shadow = kwargs.get("shadow", Gtk.ShadowType.ETCHED_OUT)
        _title: Union[Dict[str, Any], str] = kwargs.get("title", "")

        _label: RAMSTKLabel = RAMSTKLabel(_title)  # type: ignore
        _label.do_set_properties(bold=_bold)
        _label.show_all()
        self.set_label_widget(_label)

        self.set_shadow_type(_shadow)
